# -*- coding: utf-8 -*-
'''
  fastmat/Fourier.pyx
 -------------------------------------------------- part of the fastmat package

  Fourier matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .Eye cimport Eye
from .core.cmath cimport *
from .core.strides cimport *

################################################################################
################################################## class Fourier
cdef class Fourier(Matrix):

    ############################################## class properties
    # num_order - Property (read-only)
    # Return the Order of the fourier matrix.
    property order:
        def __get__(self):
            return self._order

    ############################################## class methods
    def __init__(self, order, **options):
        '''Initialize Matrix instance with a list of child matrices'''
        cdef intsize paddedSize
        cdef np.ndarray arrSamples, vecConv

        cdef bint optimize = options.get('optimize', True)
        cdef int maxStage = options.get('maxStage', 4)

        assert order >= 1, "Fourier order cannot be smaller than 1."

        # store order of Fourier
        self._order = order

        # determine if computation of the FFT via convolution is beneficial
        # first, detect the smallest size of a reasonably fast chirp-Z transform
        # start with the minimal size of 2 * N - 1
        if optimize:
            paddedSize = _findOptimalFFTSize(order * 2 - 1, maxStage)
            # check once more nothing went wrong up to this point
            assert paddedSize >= order * 2 - 1

            # now as we know where to pad to if we choose to, decide what is a
            # lesser pain to do. self._numL acts as a marker on whether to act
            # as a plain fft wrapper (== 0) or whether to use the chirp-z
            # transform in the latter case self._numL specifies the internal
            # dimension
            self._numL = (0 if (_getFFTComplexity(self.order) <
                                _getFFTComplexity(paddedSize) * 2 + paddedSize)
                          else paddedSize)

        # if we convolve, then we should prepare some stuff
        # the presence (non-None) of self._vecConvHat controls the behaviour
        # of the forward() and backward() transforms.
        if self._numL > 0:

            # create the sampling grid
            arrSamples = np.linspace(0, self.order - 1, self.order)

            # evaluate at these samples for the first numL elements
            # of the vector
            vecConv = _arrZero(1, self._numL, 1, np.NPY_COMPLEX128)
            vecConv[: self.order] = np.exp(
                +1j * (arrSamples ** 2) * np.pi / self.order)

            # now put a flipped version of the above at the very end to get
            # a circular convolution
            vecConv[self._numL - self.order + 1:] = vecConv[1:self.order][::-1]

            # get a premultiplication array for preprocessing before the
            # convolution
            self._preMult = np.exp(
                -1j * (arrSamples ** 2) * np.pi / self.order)

            # transfer function of convolution
            self._vecConvHat = np.fft.fft(vecConv)

        # set properties of matrix
        self._initProperties(self._order, self._order, np.complex128)

    ############################################## class property override
    cpdef np.ndarray _getArray(self):
        '''
        Return an explicit representation of the matrix as numpy-array.
        '''
        return np.fft.fft(np.eye(self._order, dtype=self.dtype), axis=0)

    cpdef np.ndarray _getRow(self, intsize idx):
        return self._getCol(idx)

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.sqrt(self._order)

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.sqrt(self._order)

    cpdef Matrix _getNormalized(self):
        return Fourier(self._order) * (1. / np.sqrt(self._order))

    cpdef Matrix _getGram(self):
        return Eye(self._order) * self.dtype(self._order)

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return np.exp(idxN * idxM * -2j * np.pi / self.order).astype(self.dtype)

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = _getFFTComplexity(self._order)
        return (complexity, complexity + self._order)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''
        cdef np.ndarray arrRes
        cdef STRIDE_s strResPadding
        cdef intsize mm, M = arrX.shape[1]

        if self._numL == 0:
            arrRes = np.fft.fft(arrX, axis=0)
        else:
            arrRes = _arrEmpty(
                2, self._numL, M,
                typeInfo[self._info.dtype.promote[_getFType(arrX)]].typeNum)
            strideInit(&strResPadding, arrRes, 0)
            strideSliceElements(&strResPadding, self.order, -1, 1)
            opZeroVectors(&strResPadding)

            arrRes[:self.order, :] = (self._preMult.T * arrX.T).T

            arrRes = (self._vecConvHat.T * np.fft.fft(arrRes, axis=0).T).T

            arrRes = (self._preMult.T *
                      np.fft.ifft(arrRes, axis=0)[:self.order, :].T).T

        return arrRes

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''
        cdef np.ndarray arrRes
        cdef STRIDE_s strResPadding
        cdef intsize mm, M = arrX.shape[1]

        if self._numL == 0:
            arrRes = np.fft.fft(_conjugate(arrX), axis=0)
        else:
            arrRes = _arrEmpty(
                2, self._numL, M,
                typeInfo[self._info.dtype.promote[_getFType(arrX)]].typeNum)

            strideInit(&strResPadding, arrRes, 0)
            strideSliceElements(&strResPadding, self.order, -1, 1)
            opZeroVectors(&strResPadding)

            arrRes[:self.order, :] = (self._preMult.T * _conjugate(arrX).T).T

            arrRes = (self._vecConvHat.T * np.fft.fft(arrRes, axis=0).T).T

            arrRes = (self._preMult.T *
                      np.fft.ifft(arrRes, axis=0)[:self.order, :].T).T

        _conjugateInplace(arrRes)
        return arrRes

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        return np.exp(
            np.multiply(
                *np.meshgrid(np.arange(self._order), np.arange(self._order)
                             )) * -2j * np.pi / self._order
        ).astype(self.dtype)

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # define matrix sizes and parameters
                # 35 is just any number with non-2 and non-4 primes
                # 89 is the first size for which bluestein is faster
                TEST.NUM_N      : TEST.Permutation([35, 89]),
                TEST.NUM_M      : TEST.NUM_N,
                'optimize'      : TEST.Permutation([False, True]),

                # define constructor for test instances and naming of test
                TEST.OBJECT     : Fourier,
                TEST.INITARGS   : [TEST.NUM_N],
                TEST.INITKWARGS : {'optimize': 'optimize'},
                TEST.TOL_POWER  : 3.,
                TEST.NAMINGARGS : dynFormat("%d, optimize=%d", TEST.NUM_N,
                                            'optimize'),
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Fourier(c))
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Fourier(2 ** c))
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, datatype: Fourier(2 ** c))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC, BENCH
        return DOC.SUBSECTION(
            r'Fourier transform (\texttt{fastmat.Fourier})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
The Fourier Transform realizes the mapping
\[\bm x \mapsto \bm{\mathcal{F}}_n \cdot \bm x,\]
where the Fourier matrix $\bm{\mathcal{F}}_n$
is uniquely defined by the size of the vectors it acts on.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define parameter',
                            'n = 4',
                            '',
                            '# construct the matrix',
                            'F = fm.Fourier(n)',
                            caption=r"""
This yields a Fourier $\bm{\mathcal{F}}_4$ matrix of size $4$."""),
                r"""
As a library to provide the Fast Fourier Transform we used the one provided by
NumPy \cite{four_walt2011numpy}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix $\bm{\mathcal{F}}_n$ with
$n \in \N$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED()
            ),
            DOC.BIBLIO(
                four_walt2011numpy=DOC.BIBITEM(
                    r"""
St\'efan van der Walt, S. Chris Colbert and Ga\"el Varoquaux""",
                    r"""
The NumPy Array: A Structure for Efficient Numerical Computation""",
                    r"""
Computing in Science and Engineering, Volume 13, 2011.""")
            )
        )



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/core/cmath.pxd
 -------------------------------------------------- part of the fastmat package

  fastmat striding math library. (types and funcs)

  Author      : wcw
  Introduced  : 2017-11-13
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
from libc.string cimport memcpy, memset
from libc.math cimport ceil

import numpy as np
cimport numpy as np

from .types cimport *

################################################################################
###  Basic stride operations
################################################################################
cdef void strideInit(STRIDE_s *stride, np.ndarray arr, np.uint8_t axis):
    if axis > 1:
        raise ValueError("Striding operations support 2D-arrays only.")

    stride[0].dtype         = _getFType(arr)
    if stride[0].dtype >= TYPE_INVALID:
        raise ValueError("Striding does not support given array data type.")

    stride[0].base          = arr.data                  # array base pointer
    stride[0].strideElement = arr.strides[axis]         # the selected axis
    stride[0].strideVector  = arr.strides[axis ^ 1]     # the other axis
    stride[0].numElements   = arr.shape[axis]           # array element count
    stride[0].numVectors    = arr.shape[axis ^ 1]       # vector count
    stride[0].sizeItem      = np.PyArray_ITEMSIZE(arr)  # size of one item


cdef void strideCopy(STRIDE_s *strideDst, STRIDE_s *strideSrc):
    memcpy(strideDst, strideSrc, sizeof(STRIDE_s))


cdef void strideSliceVectors(STRIDE_s *stride,
                             intsize start, intsize stop, intsize step):
    # CAUTION: NOT VERIFIED YET
    if start < 0:
        start = stride[0].numVectors

    if stop < 0:
        stop = stride[0].numVectors

    stride[0].base += start * stride[0].strideVector
    stride[0].numVectors = stop if step == 0 else (stop - start) / step
    stride[0].strideVector = stride[0].strideVector * step

cdef void strideSliceElements(STRIDE_s *stride,
                              intsize start, intsize stop, intsize step):
    if start < 0:
        start = stride[0].numElements

    if stop < 0:
        stop = stride[0].numElements

    stride[0].base += start * stride[0].strideElement
    stride[0].numElements = stop if step == 0 else (stop - start) / step
    stride[0].strideElement = stride[0].strideElement * step

cdef void strideSubgridVector(STRIDE_s *stride,
                              intsize idxVector, intsize idxElement,
                              intsize steppingElements, intsize numElements,
                              intsize steppingVectors, intsize numVectors):
    # CAUTION: NOT VERIFIED YET

    # compute new base pointer (offset subgrid stride from current one)
    stride[0].base += (stride[0].strideElement * idxElement +
                       stride[0].strideVector * idxVector)

    # now compute new spans and strides for element and vector axes. Both of
    # them originate from the vector at position (idxElement, idxVector).
    # This allows mapping butterfly operation slicing directly to a 2D stride!
    stride[0].strideVector = stride[0].strideElement * steppingVectors
    stride[0].strideElement = stride[0].strideElement * steppingElements
    stride[0].numElements = numElements
    stride[0].numVectors = numVectors

cdef void strideFlipVectors(STRIDE_s *stride):
    stride[0].base += (stride[0].numVectors - 1) * stride[0].strideVector
    stride[0].strideVector = -stride[0].strideVector

cdef void strideFlipElements(STRIDE_s *stride):
    stride[0].base += (stride[0].numElements - 1) * stride[0].strideElement
    stride[0].strideElement = -stride[0].strideElement

cdef stridePrint(STRIDE_s *stride, text=''):
    print("[%dx%d,%d] @ 0x%012X + %d * nn + %d * mm (%d Bytes) %s" %(
        stride[0].numElements, stride[0].numVectors, stride[0].dtype,
        <intsize> (stride[0].base),
        stride[0].strideElement, stride[0].strideVector, stride[0].sizeItem,
        ": %s" %(text, ) if len(text) > 0 else ""))

################################################################################
###  Operations with strides
################################################################################
cdef opCopyVector(STRIDE_s *strideDst, intsize idxVectorDst,
                  STRIDE_s *strideSrc, intsize idxVectorSrc):
    # CAUTION: NOT VERIFIED YET

    cdef intsize nn
    cdef intsize dstStride = strideDst[0].strideElement
    cdef intsize srcStride = strideSrc[0].strideElement
    cdef intsize numElements = strideDst[0].numElements
    cdef np.uint8_t sizeItem = strideDst[0].sizeItem

    # sanity check
    if ((strideDst[0].sizeItem != strideSrc[0].sizeItem or
         strideDst[0].numElements != strideSrc[0].numElements)):
        raise TypeError("Strides differ in vector length or element size.")

    # determine addresses
    cdef char *ptrSrc = (strideSrc[0].base +
                         strideSrc[0].strideVector * idxVectorSrc)
    cdef char *ptrDst = (strideDst[0].base +
                         strideDst[0].strideVector * idxVectorDst)

    if dstStride == srcStride == sizeItem:
        # element-contiguous in both strides
        memcpy(ptrDst, ptrSrc, dstStride * numElements)
    elif sizeItem == 8:
        for nn in range(numElements):
            (<np.int64_t *> ptrDst)[0] = (<np.int64_t *> ptrSrc)[0]
            ptrDst += dstStride
            ptrSrc += srcStride
    elif sizeItem == 4:
        for nn in range(numElements):
            (<np.int32_t *> ptrDst)[0] = (<np.int32_t *> ptrSrc)[0]
            ptrDst += dstStride
            ptrSrc += srcStride
    elif sizeItem == 1:
        for nn in range(numElements):
            ptrDst[0] = ptrSrc[0]
            ptrDst += dstStride
            ptrSrc += srcStride
    else:
        for nn in range(numElements):
            memcpy(ptrDst, ptrSrc, sizeItem)
            ptrDst += dstStride
            ptrSrc += srcStride


cdef opZeroVector(STRIDE_s *stride, intsize idxVector):
    # CAUTION: NOT VERIFIED YET
    cdef intsize nn
    cdef char *ptr = stride[0].base + stride[0].strideVector * idxVector
    cdef intsize strideElement = stride[0].strideElement
    cdef intsize numElements = stride[0].numElements
    cdef np.uint8_t sizeItem = stride[0].sizeItem

    if strideElement == sizeItem:
        # element-contiguous: easy memset call
        memset(ptr, 0, sizeItem * numElements)
    elif strideElement == -sizeItem:
        # flipped, but still element-contiguous: easy memset call
        memset(ptr - sizeItem * (numElements - 1), 0, sizeItem * numElements)

    # do it the long and manual way
    elif sizeItem == 8:
        for nn in range(numElements):
            (<np.int64_t *> ptr)[0] = 0
            ptr += strideElement
    elif sizeItem == 4:
        for nn in range(numElements):
            (<np.int32_t *> ptr)[0] = 0
            ptr += strideElement
    elif sizeItem == 1:
        for nn in range(numElements):
            ptr[0] = 0
            ptr += strideElement
    else:
        for nn in range(numElements):
            memset(ptr, 0, sizeItem)
            ptr += strideElement

cdef opZeroVectors(STRIDE_s *stride):
    # CAUTION: NOT VERIFIED YET
    cdef intsize nn, mm, sizeChunk
    cdef np.uint8_t sizeItem = stride[0].sizeItem
    cdef intsize numElements = stride[0].numElements
    cdef intsize numVectors = stride[0].numVectors
    cdef intsize strideElement = stride[0].strideElement
    cdef intsize strideVector = stride[0].strideVector
    cdef char *ptr = stride[0].base
    cdef char *ptrVector

    if ((strideElement == sizeItem and
         strideVector == numElements * sizeItem)):
        # both axes are contiguous: jackpot!
        memset(ptr, 0, strideVector * numVectors)
    elif strideElement == sizeItem:
        # element-contiguous: loop memset calls
        sizeChunk = numElements * sizeItem
        for mm in range(numVectors):
            memset(ptr, 0, sizeChunk)
            ptr += strideVector
    elif strideVector == sizeItem:
        # vector-contiguous: loop memset calls
        sizeChunk = numVectors * sizeItem
        for nn in range(numElements):
            memset(ptr, 0, sizeChunk)
            ptr += strideElement

    # do it the long and manual way
    elif sizeItem == 8:
        for mm in range(numVectors):
            ptrVector = ptr + strideVector
            for nn in range(numElements):
                (<np.int64_t *> ptrVector)[0] = 0
                ptrVector += strideElement
    elif sizeItem == 4:
        for mm in range(numVectors):
            ptrVector = ptr + strideVector
            for nn in range(numElements):
                (<np.int32_t *> ptrVector)[0] = 0
                ptrVector += strideElement
    elif sizeItem == 1:
        for mm in range(numVectors):
            ptrVector = ptr + strideVector
            for nn in range(numElements):
                ptrVector[0] = 0
                ptrVector += strideElement
    else:
        for mm in range(numVectors):
            ptrVector = ptr + strideVector
            for nn in range(numElements):
                memset(ptrVector, 0, sizeItem)
                ptrVector += strideElement



# -*- coding: utf-8 -*-
'''
  demo/edgeDetect.py
 -------------------------------------------------- part of the fastmat demos


  Author      : sempersn
  Introduced  :
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''

# add local fastmat directory to system path
import sys
import time

# import numpy functionality
import numpy as np
from scipy import ndimage

################################################## import modules
try:
    import matplotlib.pyplot as plt               # plotting
except ImportError:
    print("matplotlib not found. Please consider installing it to proceed.")
    sys.exit(0)

try:
    import fastmat
except ImportError:
    sys.path.append('..')
    import fastmat


print("fastmat demo: Edge detection using fast correlation")
print("---------------------------------------------------")

# read the image
arrHead = ndimage.imread("head.png", flatten=True)

# extract the dimensions
numN, numM = arrHead.shape

# create the correlation vector
c = np.zeros(numN)
c[0] = 0.5
c[1] = -0.5

# create the correlation matrix
print(" * Create the correlation matrix")
fmatLx1 = fastmat.Circulant(c)

# calculate the gradient in the directions
# using the efficient way
print(" * Perform edge detection while exploiting structure")
s = time.time()
arrEdgesX = np.abs(fmatLx1 * arrHead)
arrEdgesY = np.abs(fmatLx1 * arrHead.T).T
numFastTime = time.time() - s

# cast the matrix to a standard numpy array
print(" * Generate unstructured reference matrix for speed comparison")
fmatLx2 = fastmat.Matrix(fmatLx1.array)

# calculate the gradient in the directions
# using the inefficient way
print(" * Perform edge detection without exploting structure")
s = time.time()
arrEdgesX = np.abs(fmatLx2 * arrHead)
arrEdgesY = np.abs(fmatLx2 * arrHead.T).T
numSlowTime = time.time() - s

# calc the total gradient
arrEdges = np.sqrt(arrEdgesX**2 + arrEdgesY**2)

print("\nResults:")
print("   Fast Detection: %.3fs" %(numFastTime))
print("   Reference     : %.3fs" %(numSlowTime))
print("   Speedup Factor: %3.3f" %(numSlowTime / numFastTime))

plt.figure(1)
plt.subplot(211)
plt.imshow(arrHead, cmap=plt.cm.gray)
plt.subplot(212)
plt.imshow(arrEdges, cmap=plt.cm.gray)
plt.show()



# -*- coding: utf-8 -*-
'''
  fastmat/Eye.pyx
 -------------------------------------------------- part of the fastmat package

  Eye matrix.
ii

  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .core.types cimport *
from .core.cmath cimport _arrZero

################################################################################
################################################## class Eye
cdef class Eye(Matrix):

    ############################################## class methods
    def __init__(self, numN):
        '''
        Initialize Matrix instance with its dimensions.

        Generated a [numN x numM] matrix of small integers
        '''
        self._initProperties(numN, numN, np.int8)

    cpdef np.ndarray _getArray(Eye self):
        '''
        Return an explicit representation of the matrix as numpy-array.
        '''
        return np.eye(self.numN, dtype=self.dtype)

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        cdef np.ndarray arrRes

        arrRes = _arrZero(1, self.numN, 1, self._info.dtype[0].typeNum)
        arrRes[idx] = 1

        return arrRes

    cpdef np.ndarray _getRow(self, intsize idx):
        return self._getCol(idx)

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return 1.

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return 1.

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return 1 if (idxN == idxM) else 0

    cpdef Matrix _getNormalized(self):
        return self

    cpdef Matrix _getGram(self):
        return self

    cpdef Matrix _getT(self):
        return self

    cpdef Matrix _getH(self):
        return self

    cpdef Matrix _getConj(self):
        return self

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (0., 0.)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(Eye, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''
        return arrX

    cpdef np.ndarray _backward(Eye self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''
        return arrX

    ############################################## class reference
    cpdef np.ndarray _reference(Eye self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        return np.eye(self.numN, dtype=self.dtype)

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 35,
                TEST.NUM_M      : TEST.NUM_N,
                TEST.OBJECT     : Eye,
                TEST.INITARGS   : [TEST.NUM_N]
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Eye(c)),
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {
                BENCH.FUNC_GEN  : (lambda c: Eye(c))
            },
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Eye(2 ** c)),
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Identity Transform (\texttt{fastmat.Eye})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
For $\bm x \in \C^n$ we map \[\bm x \mapsto \bm x.\]
The identity matrix only needs the dimension $n$ of the vectors it acts on.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat',
                            '',
                            '# set the parameter',
                            'n = 10',
                            '',
                            '# construct the identity',
                            'I = fastmat.Eye(n)',
                            caption=r"""
This yields the identity matrix $\bm I_{10}$ with dimension $10$.""")
            ),
            DOC.SUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix $\bm I_n$ with $n \in \N$.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD()
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/LowRank.pyx
 -------------------------------------------------- part of the fastmat package

  Low Rank Matrix


  Author      : sempersn
  Introduced  : 2017-02-02
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Outer cimport Outer
from .Product cimport Product
from .Matrix cimport Matrix

from .Diag cimport Diag

from .core.types cimport *

################################################################################
################################################## class LowRank
cdef class LowRank(Product):

    ############################################## class properties
    # vecS - Property (read-only)
    # Return the vector of non-zero singular values entries.
    property vecS:
        def __get__(self):
            return self._vecS

    # arrU - Property (read-only)
    # Return the array of left orthogonal vectors, i.e. the image
    property arrU:
        def __get__(self):
            return self._arrU

    # arrV - Property (read-only)
    # Return the array of right orthogonal vectors, i.e.
    # the orthogonal complement of the kernel
    property arrV:
        def __get__(self):
            return self._arrV

    ########################################## Class methods
    def __init__(self, vecS, arrU, arrV):

        # complain if dimension does not match
        if vecS.ndim != 1:
            raise ValueError(
                "Singular value vector must have exactly one active dimension.")

        if arrU.ndim != 2 or arrV.ndim != 2:
            raise ValueError("Defining arrays must be 2D.")

        if vecS.shape[0] != arrU.shape[1]:
            raise ValueError("Size of U must fit to rank.")

        # determine matrix data data type from parameters, then adjust them to
        # match (parameters)
        dtype = np.promote_types(
            np.promote_types(arrU.dtype, arrV.dtype), vecS.dtype)

        # store singular values vector as copy of vecS
        # store copies of left and right orthogonal matrices
        self._arrU = arrU.astype(dtype, copy=True, subok=False)
        self._arrV = arrV.astype(dtype, copy=True, subok=False)
        self._vecS = np.atleast_1d(vecS).astype(dtype, copy=True, subok=False)

        super(LowRank, self).__init__(
            Matrix(self._arrU), Diag(vecS), Matrix(self._arrV.conj().T))

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = self.numN + self.numM + self._vecS.shape[0]
        return (complexity, complexity)

    ########################################################################
    ## forward and backward are taken from sum, product and outer
    ########################################################################

    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrU = self._arrU.astype(
            np.promote_types(self._arrU.dtype, np.float64))
        cdef np.ndarray arrV = self._arrV.astype(
            np.promote_types(self._arrV.dtype, np.float64))
        cdef np.ndarray vecS = self._vecS.astype(
            np.promote_types(self._vecS.dtype, np.float64))

        return arrU.dot(np.diag(vecS).dot(arrV.conj().T))

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                'order'         : 4,
                TEST.TOL_POWER  : (lambda param: np.sqrt(param['order'])),
                TEST.NUM_N      : 7,
                TEST.NUM_M      : TEST.Permutation([11, TEST.NUM_N]),
                'mTypeS'        : TEST.Permutation(TEST.ALLTYPES),
                'mTypeU'        : TEST.Permutation(TEST.FEWTYPES),
                'mTypeV'        : TEST.Permutation(TEST.ALLTYPES),
                'vecS'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeS',
                    TEST.SHAPE  : ('order',),
                }),
                'arrU'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeU',
                    TEST.SHAPE  : (TEST.NUM_N, 'order'),
                }),
                'arrV'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeV',
                    TEST.SHAPE  : (TEST.NUM_M, 'order'),
                }),
                TEST.INITARGS   : (lambda param : [param['vecS'](),
                                                   param['arrU'](),
                                                   param['arrV']()]),
                TEST.OBJECT     : LowRank,
                TEST.NAMINGARGS : dynFormat("%s,%s,%s", 'arrU', 'vecS', 'arrV'),
                TEST.TOL_POWER  : 3.
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH, arrTestDist
        return {
            BENCH.COMMON: {
                BENCH.FUNC_SIZE : (lambda c: 10 * c + 1),
                BENCH.FUNC_GEN  : (lambda c: LowRank(
                    arrTestDist((c + 1, ), dtype=np.float, center=2),
                    arrTestDist((10 * c + 1, c + 1), dtype=np.float, center=2),
                    arrTestDist((10 * c + 1, c + 1), dtype=np.float, center=2)))
            },
            BENCH.FORWARD: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_SIZE : (lambda c: 2 ** c),
                BENCH.FUNC_GEN  : (lambda c: LowRank(
                    arrTestDist((c, ), dtype=np.float),
                    arrTestDist((2 ** c, c), dtype=np.float),
                    arrTestDist((2 ** c, c), dtype=np.float)))
            },
            BENCH.DTYPES: {
                BENCH.FUNC_SIZE : (lambda c: 2 ** c),
                BENCH.FUNC_GEN  : (lambda c, datatype: LowRank(
                    arrTestDist((c, ), dtype=np.float),
                    arrTestDist((2 ** c, c), dtype=np.float),
                    arrTestDist((2 ** c, c), dtype=np.float)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Low Rank Matrix (\texttt{fastmat.LowRank})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
Generally one can consider the "complexity" of a matrix as the number of its
rows $n$ and columns $m$. The rank of a matrix $\bm A \in \C^{n \times m}$
always obeys the bound
    \[\Rk(\bm A) \leqslant \Min\{n,m\}.\]
If one carries out the normal matrix vector multiplication, one assumes the rank
to be essentially close to this upper bound. However if the rank of $\bm A$ is
far lower than the minimum of its dimensions, then one carries out a lot of
redundant tasks, when applying this matrix to a vector. But if one computes the
singular value decomposition (SVD) of $\bm A = \bm U \bm \Sigma \bm V^\herm$,
then one can express $\bm A$ as a sum of rank-$1$ matrices as
    \[\bm A = \Sum{i = 1}{r}{\sigma_i \bm u_{i} \bm v^\herm_{i}}.\]
If $r = \Rk $ is much smaller than the minimum of the dimensions, then one can
save a lot of computational effort in applying $\bm A$ to a vector.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# define all parameters',
                            'S = np.random.randn(2)',
                            'U = np.random.randn(20,2)',
                            'V = np.random.randn(20,2)',
                            '',
                            '# define the matrix',
                            'L = fm.LowRank(S, U, V)',
                            center=r"""
We define a matrix $\bm L = \bm U \bm S \bm V^\herm \in \R^{20 \times 20}$
with rank $2$""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix $\bm A \in \R^{n \times n}$ with
rank approximately $n/10$.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            )
        )



# -*- coding: utf-8 -*-
#cython: boundscheck=False
'''
  fastmat/MLUltraSound.py
 -------------------------------------------------- part of the fastmat package

  MLUltraSound matrix.


  Author      : sempersn
  Introduced  : 2017-10-06
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .core.types cimport *
from .core.cmath cimport *
from .Matrix cimport Matrix
from .Partial cimport Partial
from .Product cimport Product
from .Toeplitz cimport Toeplitz
from .Eye cimport Eye
from .Diag cimport Diag
from .Kron cimport Kron
from .MLToeplitz cimport MLToeplitz
from .DiagBlocks cimport DiagBlocks


################################################################################
################################################## class Toeplitz
cdef class MLUltraSound(Partial):

    ############################################## class properties
    # vecC - Property (read-only)
    # Return the matrix-defining column vector of the circulant matrix
    property tenT:
        def __get__(self):
            return self._tenT

    ############################################## class methods
    def __init__(self, tenT, **options):
        '''
        Initialize MLUltraSound Matrix instance.
        '''

        # store the defining elements and extract the needed dimensions
        self._tenT = np.copy(tenT)
        self._numBlocksN = tenT.shape[0]
        self._numSize1 = int((tenT.shape[2] + 1) /2)
        self._numSize2 = int((tenT.shape[3] + 1) /2)
        self._arrN = np.array([self._numSize1, self._numSize2], dtype='int')

        # construct one sample MLToeplitz instance to get optimal fourier
        # sizes
        _T = MLToeplitz(np.copy(tenT[0, 0, :, :]))

        # subselection index of the embedded partials
        indK = _T.indicesN.reshape((-1, 1))

        # the diagonalizing matrices of each embedded block
        F = _T._content[0]._content[-1]

        # build up the whole diagonalizing matrix
        # TODO checkout if this can be speeded up by doing FFTs
        K = Kron(
            Eye(self._numBlocksN),
            *F._content
        )

        # allocate memory for the diagonal matrix
        diags = np.empty((
            self._numBlocksN,
            self._numBlocksN,
            F.numN
        ), dtype='complex')

        # calculate the large subselection index array for the whole matrix
        rgn = np.arange(self._numBlocksN) *F.numN
        arrIndicesN = (rgn + np.repeat(
            indK, self._numBlocksN, 1
        )).reshape(-1, order='F')

        # extract the diagonalizing stuff from the nested toeplitz matrices
        for ii in range(self._numBlocksN):
            for jj in range(self._numBlocksN):
                T = MLToeplitz(np.copy(tenT[ii, jj, :, :]))
                diags[ii, jj, :] = (T._content[0]._content[1].vecD)[:]

        # construct the composing matrices
        B = DiagBlocks(diags)
        P = Product(K.H, B, K)

        # call the parent constructor
        super(MLUltraSound, self).__init__(P, N=arrIndicesN, M=arrIndicesN)

        # Currently Fourier matrices bloat everything up to complex double
        # precision, therefore make sure tenT matches the precision of the
        # matrix itself
        if self.dtype != self._tenT.dtype:
            self._tenT = self._tenT.astype(self.dtype)

    cpdef np.ndarray _getArray(self):
        '''Return an explicit representation of the matrix as numpy-array.'''
        return self._reference()

    cpdef Matrix _getNormalized(self):
        cdef intsize ii, jj

        cdef intsize stride = np.prod(self._arrN)

        cdef np.ndarray arrNorms = np.zeros(self.numM)

        for ii in range(self._numBlocksN):
            for jj in range(self._numBlocksN):
                arrNorms[jj *stride:(jj +1) *stride] += self._normalizeCore(
                    self._tenT[ii, jj, :, :]
                )

        #print(np.sqrt(arrNorms))

        return self * Diag(1 / np.sqrt(arrNorms))

    def _normalizeCore(self, tenT):
        cdef intsize ii, numS1, numS2, numS3

        cdef intsize numL = int((tenT.shape[0] + 1) /2)

        cdef intsize numD = np.array(tenT.shape).shape[0]

        cdef np.ndarray arrT, arrNorms
        if numD == 1:
            # if we are deep enough we do the normal toeplitz stuff
            arrT = tenT

            arrNorms = np.zeros(numL)

            arrNorms[0] = np.linalg.norm(arrT[:numL]) **2

            for ii in range(numL - 1):
                arrNorms[ii + 1] = arrNorms[ii] \
                    + np.abs(arrT[2 * numL - 2 - ii]) ** 2 \
                    - np.abs(arrT[numL - ii - 1]) ** 2

        else:
            numS1 = np.prod(self._arrN[-numD :])
            numS2 = np.prod(self._arrN[-(numD - 1) :])
            arrNorms = np.zeros(numS1)
            arrT = np.zeros((tenT.shape[0], numS2))

            # go deeper in recursion and get norms of blocks
            for ii in range(tenT.shape[0]):
                arrT[ii, :] = self._normalizeCore(tenT[ii])

            numS3 = arrT.shape[1]
            arrNorms[:numS3] = np.sum(arrT[:numL, :], axis=0)

            # now do blockwise subtraction and addition
            for ii in range(numL - 1):
                arrNorms[
                    (ii +1) *numS2 : (ii +2) *numS2
                ] = arrNorms[ii *numS2 : (ii +1) *numS2] + \
                    + arrT[2 * numL - 2 - ii] \
                    - arrT[numL - ii - 1]
        return arrNorms

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (0., 0.)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        arrRes = np.zeros((self.numN, self.numM), dtype=self.dtype)

        # go through all blocks and construct the corresponding
        # MLToeplitz instance by calling its reference
        for ii in range(self._numBlocksN):
            for jj in range(self._numBlocksN):
                arrRes[
                    ii * self._numSize1 * self._numSize2:
                    (ii + 1) * self._numSize1 * self._numSize2,
                    jj * self._numSize1 * self._numSize2:
                    (jj + 1) * self._numSize1 * self._numSize2
                ] = MLToeplitz(
                    np.copy(self._tenT[ii, jj, :, :])
                )._reference()

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # 35 is just any number that causes no padding
                # 41 is the first size for which bluestein is faster
                TEST.NUM_N      : 40,
                TEST.NUM_M      : TEST.NUM_N,
                'mTypeC'        : TEST.Permutation(TEST.ALLTYPES),
                'vecC'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeC',
                    TEST.SHAPE  : (2, 2, 7, 9)
                }),
                TEST.INITARGS   : (lambda param : [param['vecC']()]),
                TEST.OBJECT     : MLUltraSound,
                TEST.TOL_POWER  : 2.,
                TEST.TOL_MINEPS : _getTypeEps(np.float64)
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c:
                                   MLUltraSound(np.random.randn(
                                       c , c, 2 * c - 1, 2 * c - 1)
                                   )),
                BENCH.FUNC_SIZE : (lambda c: c ** 3),
                BENCH.FUNC_STEP : (lambda c: c + 2)
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {},
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: MLUltraSound(
                    np.random.randn(c, c, 2 * c - 1, 2 * c - 1).astype(dt)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Multilevel Ultrasound Matrix (\texttt{fastmat.MLUltraSound})',
            DOC.SUBSUBSECTION(
                'Definition and Interface',
                r"""
                This class is an implementation of matrices, which have a block
structure, where each block itself is a $2$-level Toeplitz matrix, as
implemented in \texttt{MLToeplitz}. The name originated from the fact, that
these matrices pop up, when modeling a ultrasonic pulse echo as a superposition
of hyperbolas.

As such we can
collect the unique defining elements of $\bm H$ in $\bm h \in \bm \R^{M \times
M
\times 2 N_x - 1 \times 2 N_y - 1}$ and then set
%
\[
    \bm H_{i,j} = \bm T_{([N_x, N_y], 2)}(\bm h_{i,j}).
\]
%
Moreover, we can diagonalize each $\bm H_{i,j}$.
To this end we embed each $\bm H_{i,j}$ into a $2$-level circulant matrix
%
\begin{align}\label{eff_mlultra}
    \begingroup
    \setlength\arraycolsep{3pt}
    \begin{bmatrix}
    \mathfrak{T}_2(\bm H_{0,0})   & \dots  & \mathfrak{T}_2(\bm H_{0,M-1}) \\
    \vdots        & \ddots & \vdots \\
    \mathfrak{T}_2(\bm H_{M-1,0}) & \dots  & \mathfrak{T}_2(\bm H_{M-1,M-1}) \\
    \end{bmatrix} \endgroup =
\bm K^\herm \cdot \bm D \cdot \bm K
\end{align}
%
for
$\bm F = \bm \Fs_{2 N_x - 1} \otimes \bm \Fs_{2 N_y - 1} $,
$\bm K = \bm I_{M} \otimes \bm F$
%
and
\[\bm D =
\begingroup
\setlength\arraycolsep{3pt}
\begin{bmatrix}
\diag(\bm F \Vectorize \tilde{\bm h}_{0,0}) & \dots &
\diag(\bm F \Vectorize \tilde{\bm h}_{0,M-1}) \\
\vdots & \ddots & \vdots \\
\diag(\bm F \Vectorize \tilde{\bm h}_{M-1,0}) & \dots &
\diag(\bm F \Vectorize \tilde{\bm h}_{M-1,M-1}) \\
\end{bmatrix},
\endgroup
\]
which shows, that $\bm H$ simply is a subselection of the above block
diagonalizated matrix. This is exactly, how it is implemented in \fm{}.
                """,
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# construct the',
                            '# parameters',
                            'n = 2',
                            'l = 2',
                            't = np.arange(',
                            '2 *2 *(2 *n - 1) ** l',
                            ').reshape(',
                            '(n, n, 2 *n - 1, 2 *n - 1)',
                            ')',
                            '',
                            '# construct the matrix',
                            'T = fm.MLUltraSound(t)',
                            caption=r"""
This yields
\[\bm t = \begin{bmatrix}1 & 2 & 3 \\ 4 & 5 & 6 \\ 7 & 8 & 9 \end{bmatrix}\]
\[\bm T = \left(\begin{array}{cccc}
    1 & 3 & 7 & 9 \\
    2 & 1 & 8 & 7 \\
    4 & 5 & 1 & 3 \\
    5 & 4 & 2 & 1
\end{array}\right)\]"""),
                r"""
This class depends on \texttt{Fourier}, \texttt{DiagBlocks}, \texttt{Kron},
\texttt{Product} and \texttt{Partial}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm{\mathcal{C}} \in \R^{n \times n}$ with $n \in \N$ and all
entries drawn from a  standard Gaussian distribution.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/algs/OMP.py
 -------------------------------------------------- part of the fastmat package

  Implementation of orthogonal matching pursuit (OMP) in
  fastmat.


  Author      : sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------

  TODO:
    - optimize einsum-stuff
'''
import numpy as np
import numpy.linalg as npl

from ..base import Algorithm

################################################################################
###  OMP: Sparse recovery algorithm
################################################################################


def OMP(fmatA, arrY, numK):
    '''
    Wrapper around the ISTA algrithm to allow processing of arrays of signals
        fmatA           - input system matrix
        arrY            - input data vector (measurements)
        numK            - specified sparsity order, i.e. number of iterations
                          to run
        numN,numM       - number of rows / columns of the system matrix
        numL            - number of problems to solve
        arrDiag         - array that contains column norms of fmatA
        numStrideSize   - size of strides during norm calculation
        numStrides      - number of whole strides to go through
        numPreSteps     - number of entries that do not fit in whoel strides
    '''
    if len(arrY.shape) > 2:
        raise ValueError("Only n x m arrays are supported for OMP")

    # get the number of vectors to operate on
    numN, numM, numL = fmatA.numN, fmatA.numM, arrY.shape[1]

    fmatC = fmatA.normalized

    # determine return value data type
    returnType = np.promote_types(
        np.promote_types(fmatC.dtype, arrY.dtype), np.float32)

    # temporary array to store only support entries in
    arrXtmp     = np.zeros((numK, numL), dtype=returnType)

    # initital residual is the measurement
    arrResidual = arrY.astype(returnType, copy=True)

    # list containing the support
    arrSupport  = np.empty((numK, numL),       dtype='int')

    # matrix B that contains the pseudo inverse of A restricted to the support
    arrB        = np.zeros((numK, numN, numL), dtype=returnType)

    # A restricted to the support
    arrA        = np.zeros((numN, numK, numL), dtype=returnType)

    # different helper variables
    v2       = np.empty((numN, numL), dtype=returnType)
    v2n      = np.empty((numN, numL), dtype=returnType)
    v2y      = np.empty((numL),       dtype=returnType)
    newCols  = np.empty((numN, numL), dtype=returnType)
    arrC     = np.empty((numM, numL), dtype=returnType)
    newIndex = np.empty(numL,         dtype='int')

    # iterativly build up the solution
    for ii in range(numK):

        # do the normalized correlation step
        arrC = np.abs(fmatC.backward(arrResidual))

        # pick the maximum index in each correlation array
        newIndex = np.apply_along_axis(np.argmax, 0, arrC)

        # add these to the support
        arrSupport[ii, :] = newIndex

        # get the newly picked columns of A
        newCols = fmatA.getCols(newIndex)

        # store them into the submatrix
        arrA[:, ii, :] = newCols

        # in the first step everything is simple
        if ii == 0:
            v2  = newCols
            v2n = (v2 / npl.norm(v2, axis=0) ** 2).conj()

            v2y = np.einsum('ji,ji->i', v2n, arrY)

            arrXtmp[0, :] = v2y
            arrB[0, :, :] = v2n
        else:
            v1 = np.einsum('ijk,jk->ik', arrB[:ii, :, :], newCols)

            v2 = newCols - np.einsum('ijk,jk->ik', arrA[: , :ii, :], v1)
            v2n = (v2 / npl.norm(v2, axis=0) ** 2).conj()

            v2y = np.einsum('ji,ji->i', v2n, arrY)

            arrXtmp[:ii, :] -= v2y * v1
            arrXtmp[ii , :] += v2y

            arrB[:ii, :, :] -= np.einsum('ik,jk->jik', v2n, v1)
            arrB[ii, :, :] = v2n

        # update the residual
        arrResidual -= v2y * v2

    # return the computed vector
    arrX = np.zeros((numM, numL), dtype=returnType)
    arrX[arrSupport, np.arange(numL)] = arrXtmp

    return arrX


################################################################################
###  Maintenance and Documentation
################################################################################

################################################## inspection interface
class OMPinspect(Algorithm):

    @staticmethod
    def _getTest():
        from ..inspect import TEST, dynFormat, arrSparseTestDist
        from ..Hadamard import Hadamard
        from ..Matrix import Matrix

        def testOmp(test):
            # prepare vectors
            numM = test[TEST.NUM_M]
            test[TEST.RESULT_REF]       = np.hstack(
                [arrSparseTestDist((numM, 1), dtype=test[TEST.DATATYPE],
                                   density=1. * test['numK'] / numM).todense()
                 for nn in range(test[TEST.DATACOLS])])
            test[TEST.RESULT_INPUT]     = (test[TEST.INSTANCE] *
                                           test[TEST.RESULT_REF])
            test[TEST.RESULT_OUTPUT]    = OMP(test[TEST.INSTANCE],
                                              test[TEST.RESULT_INPUT],
                                              test['numK'])

        return {
            TEST.ALGORITHM: {
                'order'         : 5,
                TEST.NUM_N      : (lambda param: 2 ** param['order']),
                TEST.NUM_M      : TEST.NUM_N,
                'numK'          : 5,
                'typeA'         : TEST.Permutation(TEST.ALLTYPES),

                TEST.OBJECT     : Matrix,
                TEST.DATAALIGN  : TEST.ALIGNMENT.DONTCARE,
                TEST.INITARGS   : (lambda param: [
                    Hadamard(param.order).array.astype(param['typeA'])]),

                TEST.INIT_VARIANT : TEST.IgnoreFunc(testOmp),

                'strTypeA'      : (lambda param: TEST.TYPENAME[param['typeA']]),
                TEST.NAMINGARGS : dynFormat("Hadamard(%s,%s)",
                                            'order', 'strTypeA'),

                # matrix inversion always expands data type to floating-point
                TEST.TYPE_PROMOTION : np.float32
            },
        }

    @staticmethod
    def _getBenchmark():
        from ..inspect import BENCH, arrTestDist
        from ..Matrix import Matrix
        from ..Fourier import Fourier
        from ..Product import Product
        from scipy import sparse as sps

        def createTarget(M, datatype):
            '''Create test target for algorithm performance evaluation.'''

            if M < 10:
                raise ValueError("Problem size too small for OMP benchmark")

            # assume a 1:5 ratio of measurements and problem size
            # assume a sparsity of half the number of measurements
            N = int(np.round(M / 5.0))
            K = int(N / 2)

            # generate matA = [random measurement matrix] * [Fourier dictionary]
            matA = Product(Matrix(arrTestDist((N, M), datatype)), Fourier(M))

            # generate attb from random baseline support (RHS)
            arrB = matA.forward(
                sps.rand(M, 1, 1.0 * K / M).todense().astype(datatype))

            return (OMP, [matA, arrB, K])

        return {
            BENCH.COMMON: {
                BENCH.NAME      : 'OMP Algorithm',
                BENCH.DOCU      : r"""We use $\bm A = \bm M \cdot \bm \Fs$,
                    where $\bm M$ was drawn from a standard Gaussian
                    distribution and $\bm \Fs$ is a Fourier matrix. The vector
                    $\bm b \in \C^m$ of equation \eqref{omp_problem} is
                    generated from multiplying $\bm A$ with a sparse vector
                    $\bm x$.""",
                BENCH.FUNC_GEN  : (lambda c: createTarget(10 * c, np.float64)),
                BENCH.FUNC_SIZE : (lambda c: 10 * c)
            },
            BENCH.PERFORMANCE: {
                BENCH.CAPTION   : 'OMP performance'
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: createTarget(10 * c, dt)),
                BENCH.FUNC_SIZE : (lambda c: 10 * c),
                BENCH.FUNC_STEP : (lambda c: c * 10 ** (1. / 12)),
            }
        }

    @staticmethod
    def _getDocumentation():
        from ..inspect import DOC
        return DOC.SUBSECTION(
            r'Orthogonal matching Pursuit (OMP) (\texttt{fastmat.algs.OMP})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
For a given matrix $\bm A \in \C^{m \times N}$ with $m \ll N$ and a
vector $\bm b \in \C^m$ we approximately solve

\begin{align}\label{omp_problem}
    \Min\limits_{\bm x \in \C^N}\Norm{\bm x}_0 \Text{s.t.} \bm A \cdot
    \bm x = \bm x.
\end{align}

If it holds that $\bm b = \bm A \cdot \bm x_0$ for some $k$-sparse $\bm x_0$ and
$k$ is low enough, we can recover $\bm x_0$ via OMP \cite{omp_mallat1952omp}.

\begin{itemize}
\item \textbf{Input:} Full rank matrix $\bm A$, measurements $\bm b$, desired
    sparsity order $k$ of the reconstruction
\item \textbf{Output:} Reconstruction vector $\bm x$.
\end{itemize}

This type of problem as the one described above occurs in Compressed Sensing
and Sparse Signal Recovery, where signals are approximated by sparse
representations.""",
                DOC.SNIPPET('# import the packages',
                            'import numpy.linalg as npl',
                            'import numpy as np',
                            'import fastmat as fm',
                            'import fastmat.algs as fma',
                            '',
                            '# define the dimensions',
                            '# and the sparsity',
                            'n, k = 512, 3',
                            '',
                            '# define the sampling positions',
                            't = np.linspace(0, 20 * np.pi, n)',
                            '',
                            '# construct the convolution matrix',
                            'c = np.cos(2 * t)',
                            'C = fm.Circulant(c)',
                            '',
                            '# create the ground truth',
                            'x = np.zeros(n)',
                            'x[npr.choice(range(n),',
                            '             k, replace=0)] = 1',
                            'b = C * x',
                            '',
                            '# reconstruct it',
                            'y = fma.OMP(C, b, k)',
                            '',
                            '# test if they are close in the',
                            '# domain of C',
                            'print(npl.norm(C * y - b))',
                            caption=r"""
We describe a sparse deconvolution problem, where the signal is in $\R^{512}$
and consists of $3$ windowed osine pulses of the form $\bm c$ with circulant
displacement. Then we take the convolution and try to recover the location of
the pulses using the OMP algorithm."""),
                r"""
\textit{Hint:} The algorithm exploits two mathematical shortcuts. First it
obviously uses the fast transform of the involved system matrix during the
correlation step and second it uses a method to calculate the pseudo
inverse after a rank-$1$ update of the matrix."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
We use $\bm A = \bm M \cdot \bm \Fs$, where $\bm M$ was drawn from a standard
Gaussian distribution and $\bm \Fs$ is a Fourier matrix. The vector $\bm b \in
\C^m$ of equation \eqref{ista_lasso} is generated from multiplying $\bm A$ with
a sparse vector $\bm x$.""",
                DOC.PLOTPERFORMANCE(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            ),
            DOC.BIBLIO(
                omp_mallat1952omp=DOC.BIBITEM(
                    r'S. G. Mallat and Zhifeng Zhang',
                    r'Matching pursuits with time-frequency dictionaries',
                    r"""
IEEE Transactions on Signal Processing, vol. 41, no. 12, pp. 3397-3415,
Dec 1993""")
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/core/calibration.py
 -------------------------------------------------- part of the fastmat package

  Provides generastion and handling of calibration values for estimating the
  runtime properties of transforms.


  Author      : wcw
  Introduced  : 2017-08-15
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
from ..Matrix import MatrixCalibration

# stores calibration data as tuple (offsetFwd/Bwd, gainFwd/Bwd) over class
calData = {}


def saveCalibration(filename):
    import json
    import os
    outData = {target.__name__: cal.export() for target, cal in calData.items()}

    # force existance of output file
    dirname=os.path.dirname(filename)
    if dirname != '' and not os.path.exists(dirname):
        os.makedirs(dirname)

    with open(filename, 'w') as f:
        string = json.dumps(outData)
        f.write(string.replace('],', '],' + os.linesep) + os.linesep)


def loadCalibration(filename):
    import json
    from .. import classes
    classNames = {item.__name__: item for item in classes}

    with open(filename, 'r') as f:
        inData = json.load(f)

    calData.clear()
    for name, data in inData.items():
        item = classNames.get(name, None)
        if item is not None:
            calData[item] = MatrixCalibration(*data)


def getMatrixCalibration(target):
    return calData.get(target, None)


def calibrateClass(target, **options):
    from fastmat.inspect import Benchmark, BENCH
    from .. import flags

    calOptions = {
        BENCH.LIMIT_ITER    : 0.1,
        BENCH.LIMIT_INIT    : 0.1,
        BENCH.LIMIT_SIZE    : 1e6,
        BENCH.LIMIT_MEMORY  : 1e5,
        BENCH.STEP_MINITEMS : 3,
        BENCH.MEAS_MINTIME  : 3e-3,
        BENCH.MEAS_MINRUNS  : 3,
        BENCH.MEAS_MINREPS  : 3,
        BENCH.FUNC_STEP     : lambda x: x + 1,
        'verbose'           : False
    }
    calOptions.update(options)

    # allow the function to skip estimation of calibration values and return
    # benchmark results that were generated with parameters matching the ones
    # that were used for the calibration
    benchmarkOnly = options.get('benchmarkOnly', False)
    if not benchmarkOnly:
        _bypassAllow = flags.bypassAllow
        flags.bypassAllow = False

    B = Benchmark(target)
    B.verbosity = calOptions.get('verbose', False)
    B.run('overhead', **calOptions)

    # transform bypass deactivated when calibrating
    if benchmarkOnly:
        return B
    else:
        flags.bypassAllow = _bypassAllow

    def estimate(nameTime, nameComplexity, nameOvhNested, nameEffNested):
        arr = np.nan_to_num(B.getResult('overhead', 'numN',
                                        nameTime, nameComplexity,
                                        nameOvhNested, nameEffNested))

        # remove runtimes of nested classes from the result
        arrTime             = arr[:, 1]
        arrComplexity       = arr[:, 2]
        arrOverheadNested   = arr[:, 3]
        arrEffortNested     = arr[:, 4]
        arrNested           = np.nan_to_num(arrOverheadNested + arrEffortNested)

        arrTime -= arrNested

        # keep entries with strictly positive runtimes and positive complexity
        arrIter         = arr[(arrTime > 0.) & (arrComplexity >= 0.), 1:3]
        arrTimeIter     = arrIter[:, 0]
        arrCplxIter     = arrIter[:, 1]

        offset, gain = 0., 0.
        if arrIter.shape[0] > 0:
            # estimate gain
            cpxRange = np.ptp(arrCplxIter)
            gain = (np.ptp(arrTimeIter) / cpxRange if cpxRange > 0. else 0.)

            # estimate offset iteratively
            for nn in range(20):
                delta = arrTimeIter - gain * arrCplxIter
                delta = delta[delta > -offset]
                if delta.shape[0] < 3:
                    break

                dOffset = (0 if delta.shape[0] < 1 else np.median(delta))
                offset += dOffset
                arrTimeIter -= dOffset

                if dOffset < offset * 1e-2:
                    break

        return offset, gain

    ovhForward, effForward = estimate('forwardMin',
                                      BENCH.RESULT_COMPLEXITY_F,
                                      BENCH.RESULT_OVH_NESTED_F,
                                      BENCH.RESULT_EFF_NESTED_F)

    ovhBackward, effBackward = estimate('backwardMin',
                                        BENCH.RESULT_COMPLEXITY_B,
                                        BENCH.RESULT_OVH_NESTED_B,
                                        BENCH.RESULT_EFF_NESTED_B)

    cal = MatrixCalibration(ovhForward, ovhBackward, effForward, effBackward)
    calData[target] = cal

    return cal, B



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/Hadamard.pyx
 -------------------------------------------------- part of the fastmat package

  Hadamard matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
#from libc.stdlib cimport malloc, free
#from libc.string cimport memcpy

import numpy as np
cimport numpy as np

from .core.types cimport nptype, _getNpType
from .core.strides cimport *
from .core.cmath cimport _arrEmpty
from .Matrix cimport Matrix
from .Eye cimport Eye

# have a very lazy import to avoid initialization of scipy.linalg during import
# of main module
spHadamard = None


cdef void _hadamardCore(STRIDE_s *strA, STRIDE_s *strB, TYPE_IN dummy):
    cdef intsize vv, ee

    cdef char *ptrVectorA
    cdef char *ptrElementA
    cdef char *ptrVectorB
    cdef char *ptrElementB

    cdef TYPE_IN a, b

    ptrVectorA = strA.base
    ptrVectorB = strB.base
    for vv in range(strA[0].numVectors):
        ptrElementA = ptrVectorA
        ptrElementB = ptrVectorB
        for ee in range(strA[0].numElements):
            a = (<TYPE_IN *> ptrElementA)[0]
            b = (<TYPE_IN *> ptrElementB)[0]
            (<TYPE_IN *> ptrElementA)[0] = a + b
            (<TYPE_IN *> ptrElementB)[0] = a - b

            ptrElementA += strA.strideElement
            ptrElementB += strB.strideElement

        ptrVectorA += strA.strideVector
        ptrVectorB += strB.strideVector


################################################################################
################################################## class Hadamard
cdef class Hadamard(Matrix):

    ############################################## class properties
    # order - Property (read-only)
    # Return the order of the hadamard matrix.
    property order:
        def __get__(self):
            return self._order

    ############################################## class methods
    def __init__(self, order):
        '''Initialize Matrix instance'''
        if order < 1:
            raise ValueError("Hadamrd: Order must be larger than 0.")

        self._order = order

        # set properties of matrix
        numN = 2 ** self._order
        self._initProperties(numN, numN, np.int8, cythonCall=True,
                             forceInputAlignment=True)

    cpdef np.ndarray _getArray(self):
        '''
        Return an explicit representation of the matrix as numpy-array.
        '''
        return self._reference()

    ############################################## class property override
    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.sqrt(self.numN)

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return np.sqrt(self.numN)

    cpdef Matrix _getNormalized(self):
        return Hadamard(self.order) * np.float32(1. / np.sqrt(self.numN))

    cpdef Matrix _getGram(self):
        return Eye(self.numN) * np.float32(self.numN)

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = self.numN * self.order
        return (complexity, complexity + 1)

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''
        Calculate the forward transform of this matrix.
        '''
        cdef nptype dtype = typeInfo[typeRes].typeNum
        cdef intsize N = arrX.shape[0], M = arrX.shape[1], order = self._order
        cdef intsize mm, oo
        cdef STRIDE_s strInput, strOutput, strA, strB
        cdef intsize butterflyDistance, butterflyCount

        strideInit(&strInput, arrX, 0)
        strideInit(&strOutput, arrRes, 0)

        for mm in range(M):
            opCopyVector(&strOutput, mm, &strInput, mm)

            butterflyDistance = 1
            butterflyCount = N / 2
            for oo in range(order):
                strideCopy(&strA, &strOutput)
                strideCopy(&strB, &strOutput)

                # Butterfly map over iterations of oo
                #        oo |0 |1  |2
                # element 0 |A |A  |A
                # element 1 |A | B | B
                # element 2 |B |A  |  C
                # element 3 |B | B |   D
                # element 4 |C |C  |A
                # element 5 |C | D | B
                # element 6 |D |C  |  C
                # element 7 |D |D |   D
                strideSubgridVector(&strA, mm, 0,
                                    1, butterflyDistance,
                                    2 * butterflyDistance, butterflyCount)
                strideSubgridVector(&strB, mm, butterflyDistance,
                                    1, butterflyDistance,
                                    2 * butterflyDistance, butterflyCount)

                if typeX == TYPE_FLOAT32:
                    _hadamardCore[np.float32_t](&strA, &strB, 0)
                elif typeX == TYPE_FLOAT64:
                    _hadamardCore[np.float64_t](&strA, &strB, 0)
                elif typeX == TYPE_COMPLEX64:
                    _hadamardCore[np.complex64_t](&strA, &strB, 0)
                elif typeX == TYPE_COMPLEX128:
                    _hadamardCore[np.complex128_t](&strA, &strB, 0)
                elif typeX == TYPE_INT64:
                    _hadamardCore[np.int64_t](&strA, &strB, 0)
                elif typeX == TYPE_INT32:
                    _hadamardCore[np.int32_t](&strA, &strB, 0)
                elif typeX == TYPE_INT8:
                    _hadamardCore[np.int8_t](&strA, &strB, 0)
                else:
                    raise NotImplementedError(
                        "Hadamard: %d not supported." %(typeX))

                butterflyDistance <<= 1
                butterflyCount >>= 1

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        '''
        Calculate the backward transform of this matrix.
        '''
        return self._forwardC(arrX, arrRes, typeX, typeRes)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        global spHadamard
        if spHadamard is None:
            spHadamard = __import__('scipy.linalg', globals(), locals(),
                                    ['hadamard']).hadamard

        return spHadamard(self.numN, dtype=self.dtype)

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # define matrix sizes and parameters
                'order'         : TEST.Permutation([4, 6]),
                TEST.NUM_N      : (lambda param : 2 ** param['order']),
                TEST.NUM_M      : TEST.NUM_N,

                # define constructor for test instances and naming of test
                TEST.OBJECT     : Hadamard,
                TEST.INITARGS   : ['order'],
                TEST.NAMINGARGS : dynFormat("%d", 'order')
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Hadamard(c)),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c),
                BENCH.FUNC_STEP : (lambda c: c + 1),
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {},
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, datatype: Hadamard(c))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Hadamard Matrix (\texttt{fastmat.Hadamard})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
A Hadamard Matrix is recursively defined as
    \[\bm H_n = \bm H_1 \otimes \bm H_{n-1},\]
where
    \[\bm H_1 = \left(\begin{array}{cc} 1 & 1 \\ 1 & -1 \end{array}\right)\]
and $\bm H_0 = (1)$. Obviously the dimension of $\bm H_n$ is $2^n$. The
transform is realized with the Fast Hadamard Transform (FHT).""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the parameter',
                            'n = 4',
                            '',
                            '# construct the matrix',
                            'H = fm.Hadamard(n)',
                            caption=r"""
This yields a Hadamard matrix $\bm{\mathcal{H}}_4$ of order $4$,
 i.e. with $16$ rows and columns."""),
                r"""
The algorithm we used is described in \cite{hada_hershey1997hadamard} and was
implemented in Cython \cite{hada_smith2011cython}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm \Hs_k$ and $n = 2^k$ with $n, k \in \N$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            ),
            DOC.BIBLIO(
                hada_hershey1997hadamard=DOC.BIBITEM(
                    r'Rao K. Yarlagadda, John E. Hershey',
                    r"""
Hadamard Matrix Analysis and Synthesis, With Applications to Communications
and Signal/Image Processing""",
                    r"""
The Springer International Series in Engineering and Computer Science,
Volume 383, 1997"""),
                hada_smith2011cython=DOC.BIBITEM(
                    r"""
Stefan Behnel, Robert Bradshaw, Craig Citro, Lisandro Dalcin,
Dag Sverre Seljebotn and Kurt Smith""",
                    r'Cython: The Best of Both Worlds',
                    r'Computing in Science and Engineering, Volume 13,2011.')
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/Sparse.py
 -------------------------------------------------- part of the fastmat package

  Sparse matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
cimport numpy as np
import numpy as np
from scipy.sparse import spmatrix

from .Matrix cimport Matrix
from .core.types cimport *
from .core.cmath cimport _conjugate

################################################################################
################################################## class Sparse
cdef class Sparse(Matrix):

    ############################################## class properties
    # spArray, spArrayH - Property (read-only)
    # Return the scipy sparse matrix or its hermitian transpose.
    property spArray:
        def __get__(self):
            return self._spArray

    property spArrayH:
        def __get__(self):
            if self._spArrayH is None:
                self._spArrayH = self._spArray.T.conj().tocsr()
            return self._spArrayH

    ############################################## class methods
    def __init__(self, value):
        '''Initialize Matrix instance'''
        if not isinstance(value, spmatrix):
            raise TypeError("Sparse: Use Matrix() for numpy ndarrays."
                            if isinstance(value, np.ndarray)
                            else "Sparse: Matrix is not a scipy spmatrix")

        self._spArray = value.tocsr()

        # set properties of matrix
        self._initProperties(self._spArray.shape[0], self._spArray.shape[1],
                             self._spArray.dtype)

    cpdef np.ndarray _getArray(self):
        '''
        Return an explicit representation of the matrix as numpy-array.
        '''
        return self._spArray.todense()

    ############################################## class property override
    cpdef np.ndarray _getCol(self, intsize idx):
        return np.squeeze(self._spArray.getcol(idx).todense())

    cpdef np.ndarray _getRow(self, intsize idx):
        return np.squeeze(_conjugate(self.spArrayH.getcol(idx).todense()))

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self._spArray[idxN, idxM]

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = self._spArray.getnnz()
        return (complexity, complexity)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''
        return self._spArray.dot(arrX)

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''
        return self.spArrayH.dot(arrX)

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        return self._spArray.todense()

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat, arrSparseTestDist
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 25,
                TEST.NUM_M      : TEST.Permutation([30, TEST.NUM_N]),
                'mType'         : TEST.Permutation(TEST.ALLTYPES),
                'density'       : .1,
                TEST.INITARGS   : (lambda param: [
                    arrSparseTestDist(
                        (param[TEST.NUM_N], param[TEST.NUM_M]), param['mType'],
                        density=param['density'], compactFullyOccupied=True
                    )]),
                TEST.OBJECT     : Sparse,
                'strType'       : (lambda param: TEST.TYPENAME[param['mType']]),
                TEST.NAMINGARGS : dynFormat("%s,%s", 'strType', TEST.NUM_M)
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        import scipy.sparse as sps
        return {
            BENCH.FORWARD: {
                BENCH.FUNC_GEN  : (lambda c: Sparse(
                    sps.rand(c, c, 0.1, format='csr')))
            },
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Sparse(
                    sps.rand(2 ** c, 2 ** c, 0.1, format='csr')))
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: (
                    Sparse(sps.rand(2 ** c, 2 ** c, 0.1,
                                    format='csr', dtype=dt))
                    if np.issubdtype(dt, np.float) else None))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Sparse Matrix (\texttt{fastmat.Sparse})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
\[x \mapsto \bm S x,\]
where $\bm S$ is a \texttt{scipy.sparse} matrix.
To provide a high level of generality, the user has to make use of the standard
\texttt{scipy.sparse} matrix constructors and pass them to \fm{} during
construction. After that a \texttt{Sparse} matrix can be used like every other
type in \fm{}.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# import scipy to get',
                            '# the constructor',
                            'import scipy.sparse.rand as r',
                            '',
                            '# set the matrix size',
                            'n = 100',
                            '',
                            '# construct the sparse matrix',
                            'S = fm.Sparse(',
                            '        r(',
                            '            n,',
                            '            n,',
                            '            0.01,',
                            "            format='csr'",
                            '        ))',
                            caption=r"""
This yields a random sparse matrix with 1\% of its entries occupied drawn from
a random distribution."""),
                r"""
It is also possible to directly cast SciPy sparse matrices into the \fm{} sparse
matrix format as follows.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# import scipy to get',
                            '# the constructor',
                            'import scipy.sparse as ss',
                            '',
                            '# construct the SciPy sparse matrix',
                            'S_scipy = ss.csr_matrix(',
                            '        [',
                            '            [1, 0, 0],',
                            '            [1, 0, 0],',
                            '            [0, 0, 1]',
                            '        ]',
                            '    )',
                            '',
                            '# construct the fastmat sparse matrix',
                            'S = fm.Sparse(S_scipy)'),
                r"""
\textit{Hint:} The \texttt{format} specifier drastically influences performance
during multiplication of these matrices. From our experience \texttt{'csr'}
works best in these cases.

For this matrix class we used the already tried and tested routines of SciPy
\cite{srs_walt2011numpy}, so we merely provide a convenient wrapper to integrate
nicely into \fm{}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a sparse matrix
$\bm S \in \R^{n \times n}$ with $10$\% of non-zero entries""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(typelist=['f32', 'f64']),
                DOC.PLOTTYPEMEMORY(typelist=['f32', 'f64'])
            ),
            DOC.BIBLIO(
                srs_walt2011numpy=DOC.BIBITEM(
                    r"""
St\'efan van der Walt, S. Chris Colbert and Ga\"el Varoquaux""",
                    r"""
The NumPy Array: A Structure for Efficient Numerical Computation""",
                    r"""
Computing in Science and Engineering, Volume 13, 2011.""")
            )
        )



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/Circulant.py
 -------------------------------------------------- part of the fastmat package

  Circulant matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------

  TODO:
    - sort out when optimizations are possible (real/complex, such stuff)
    - specify when this transform was introduced
'''
import numpy as np
cimport numpy as np

from .core.types cimport *
from .core.cmath cimport *
from .Matrix cimport Matrix
from .Partial cimport Partial
from .Product cimport Product
from .Fourier cimport Fourier
from .Diag cimport Diag


################################################################################
################################################## class Circulant
cdef class Circulant(Partial):

    ############################################## class properties
    # vecC - Property (read-only)
    # Return the matrix-defining column vector of the circulant matrix
    property vecC:
        def __get__(self):
            return self._vecC

    ############################################## class methods
    def __init__(self, vecC, **options):
        '''
        Initialize Circulant Matrix instance.

        Circulant([ d c C C C]) represents this matrix:
          [ d C C C c ]
          [ c d C C C ]
          [ C c d C C ]
          [ C C c d C ]
          [ C C C c d ]

        the generating vector passed as vecC may be zero-padded to increase
        computation efficiency of involved fft operations. However, the padded
        zeros must be appended to the center of the generating column. The
        generating vector must head the column vector to define the resulting
        lower triangular matrix. The upper triangular is defined by a reversed
        copy of the vector at the end of column vector. This is short by one
        entry, which is already placed at the very first position. The minimal
        column vector length is one element short of twice the generating vector
        length. Thus, padding is only efficient for really bad behaved fourier
        transforms.
          [ d c C C C 0 0 0 0 0 0 0 c C C C]

        Valid options
        -------------
         'pad'=[FALSE true]
            perform zero-padding of Circulant for efficiency. By default set to
            False as padding introduces significant overhead to non-padded for
            most but some sizes.

        All options specified will also be passed on to the generation of the
        underlying Product instance initialization.
        '''

        # save generating vector. Matrix sizes will be set by Product
        # during its intitalization (invoked by 'super' below)
        self._vecC = np.atleast_1d(np.squeeze(np.copy(vecC)))

        assert self._vecC.ndim == 1, "Column-definition vector must be 1D."
        assert len(self._vecC) >= 1, "Vector must have at least one entry"

        cdef bint optimize = options.get('optimize', True)
        cdef int maxStage = options.get('maxStage', 4)

        cdef intsize size = len(vecC)
        cdef intsize paddedSize, minimalSize = size * 2 - 1
        cdef np.ndarray arrIndices

        # determine if zero-padding of the convolution to achieve a better FFT
        # size is beneficial or not
        if optimize:
            paddedSize = _findOptimalFFTSize(minimalSize, maxStage)

            assert paddedSize >= size * 2 - 1

            if _getFFTComplexity(size) > _getFFTComplexity(paddedSize):
                # zero-padding pays off, so do it!
                vecC = np.concatenate([vecC,
                                       np.zeros((paddedSize - minimalSize,),
                                                dtype=vecC.dtype),
                                       vecC[1:]])
                size = paddedSize

        # Describe circulant matrix as product of data and vector in fourier
        # domain. Both fourier matrices cause scaling of the data vector by
        # size, which will be compensated in Diag().

        # Create inner product
        cdef Fourier FN = Fourier(size)
        cdef Product P = Product(FN.H, Diag(np.fft.fft(vecC) / size),
                                 FN, **options)

        # initialize Partial of Product. Only use Partial when padding size
        if size == len(self._vecC):
            super(Circulant, self).__init__(P)
        else:
            # generate index array once to save memory by one shared reference
            arrIndices = np.arange(len(self._vecC))
            super(Circulant, self).__init__(P, N=arrIndices, M=arrIndices)

        # Currently Fourier matrices bloat everything up to complex double
        # precision, therefore make sure vecC matches the precision of the
        # matrix itself
        if self.dtype != self._vecC.dtype:
            self._vecC = self._vecC.astype(self.dtype)

    cpdef np.ndarray _getArray(self):
        '''Return an explicit representation of the matrix as numpy-array.'''
        return self._reference()

    ############################################## class property override
    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self._vecC[(idxN - idxM) % self.numN]

    cpdef np.ndarray _getCol(self, intsize idx):
        '''Return selected columns of self.array'''
        cdef np.ndarray arrRes = _arrEmpty(
            1, self.numN, 0, self._info.dtype[0].typeNum)
        self._roll(arrRes, idx)
        return arrRes

    cpdef np.ndarray _getRow(self, intsize idx):
        '''Return selected rows of self.array'''
        cdef np.ndarray arrRes = _arrEmpty(
            1, self.numN, 0, self._info.dtype[0].typeNum)
        self._roll(arrRes[::-1], self.numN - idx - 1)
        return arrRes

    cpdef Matrix _getGram(self):
        cdef Fourier F = self.content[0].content[2]
        return Circulant(np.fft.ifft(abs(np.fft.fft(self.vecC)) ** 2))

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (0., 0.)

    ############################################## internal roll core
    cdef void _roll(self, np.ndarray vecOut, intsize shift):
        '''Return self.vecC rolled by 'shift' elements.'''
        if shift == 0:
            vecOut[:] = self._vecC
        else:
            vecOut[:shift] = self._vecC[self.numN - shift:]
            vecOut[shift:] = self._vecC[:self.numN - shift]

    cpdef Matrix _getNormalized(self):
        return self * (1. / np.linalg.norm(self._vecC))

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrRes
        cdef intsize ii, N = self.numN, M = self.numM

        arrRes = np.empty((N, M), dtype=self.dtype)
        arrRes[:, 0] = self._vecC
        for ii in range(N):
            self._roll(arrRes[:, ii], ii)

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # 35 is just any number that causes no padding
                # 41 is the first size for which bluestein is faster
                TEST.NUM_N      : TEST.Permutation([31, 41]),
                TEST.NUM_M      : TEST.NUM_N,
                'mTypeC'        : TEST.Permutation(TEST.ALLTYPES),
                'optimize'      : True,
                TEST.PARAMALIGN : TEST.Permutation(TEST.ALLALIGNMENTS),
                'vecC'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeC',
                    TEST.SHAPE  : (TEST.NUM_N, ),
                    TEST.ALIGN  : TEST.PARAMALIGN
                }),
                TEST.INITARGS   : (lambda param : [param['vecC']()]),
                TEST.INITKWARGS : {'optimize' : 'optimize'},
                TEST.OBJECT     : Circulant,
                TEST.NAMINGARGS : dynFormat("%s,optimize=%s",
                                            'vecC', str('optimize')),
                TEST.TOL_POWER  : 2.,
                TEST.TOL_MINEPS : _getTypeEps(np.float64)
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c:
                                   Circulant(np.random.randn(2 ** c))),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c),
                BENCH.FUNC_STEP : (lambda c: c + 1),
            },
            BENCH.FORWARD: {
                BENCH.FUNC_GEN  : (lambda c: Circulant(np.random.randn(c),
                                                       pad=True)),
                BENCH.FUNC_SIZE : (lambda c: c),
                BENCH.FUNC_STEP : (lambda c: c * 10 ** (1. / 12))
            },
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {},
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: Circulant(
                    np.random.randn(2 ** c).astype(dt)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Circulant Matrix (\texttt{fastmat.Circulant})',
            DOC.SUBSUBSECTION(
                'Definition and Interface',
                r"""
Circulant matrices realize the following mapping
\[\bm x \mapsto \bm C \cdot \bm x = \bm c * \bm x,\]
with $\bm x \in \C^n$ and
\[\bm C = \left(\begin{array}{cccc}
    c_1     & c_n       & \dots     & c_2       \\
    c_2     & c_1       & \ddots    & c_3       \\
    \vdots  & \vdots    & \ddots    & \vdots    \\
    c_n     & c_{n-1}   & \dots     & c_1
\end{array}\right).\]
This means that $\bm C$ is completely defined by its first column and realizes
the convolution with the vector $\bm c$.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            'import numpy as np',
                            '',
                            '# construct the',
                            '# parameter',
                            'n = 4',
                            'c = np.array([1, 0, 3, 6])',
                            '',
                            '# construct the matrix',
                            'C = fm.Circulant(c)',
                            caption=r"""
This yields
\[\bm c = (1,0,3,6)^T\]
\[\bm C = \left(\begin{array}{cccc}
    1 & 6 & 3 & 0 \\
    0 & 1 & 6 & 3 \\
    3 & 0 & 1 & 6 \\
    6 & 3 & 0 & 1
\end{array}\right)\]"""),
                r"""
This class depends on \texttt{Fourier}, \texttt{Diag}, \texttt{Product} and
\texttt{Partial}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm{\mathcal{C}} \in \R^{n \times n}$ with $n \in \N$ and first columns
entries drawn from a  standard Gaussian distribution.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/inspect/test.py
 -------------------------------------------------- part of the fastmat package

  Implementations for class testing (quality management).


  Author      : wcw
  Introduced  : 2017-07-10
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import itertools
import numpy as np
from pprint import pprint

from .common import *
from ..core.types import _getTypeEps, safeTypeExpansion
from ..Matrix import Matrix
from ..Diag import Diag
from ..Eye import Eye
from ..Parametric import Parametric
from ..Zero import Zero


################################################## TEST CONSTANT-def class
class TEST(NAME):
    CLASS           = 'class'
    TRANSFORMS      = 'transform'
    ALGORITHM       = 'algorithm'

    OBJECT          = 'object'
    ARGS            = 'testArgs'
    KWARGS          = 'testKwargs'
    NUM_N           = 'numN'
    NUM_M           = 'numM'
    DATACOLS        = 'numCols'
    DATATYPE        = 'dataType'
    DATASHAPE       = 'dataShape'
    DATAALIGN       = 'dataAlign'
    DATASHAPE_T     = 'dataShapeBackward'
    DATAGEN         = 'dataGenerator'
    DATACENTER      = 'dataDistCenter'
    DATAARRAY       = 'arrData'
    INIT            = 'init'
    INITARGS        = 'args'
    INITKWARGS      = 'kwargs'
    INIT_VARIANT    = 'initVariant'
    INSTANCE        = 'instance'
    REFERENCE       = 'reference'
    QUERY           = 'query'
    NAMINGARGS      = 'namingArgs'
    NAMING          = 'naming'
    TOL_MINEPS      = 'tolMinEps'
    TOL_POWER       = 'tolPower'
    IGNORE          = 'ignore'
    PARAMALIGN      = 'alignment'
    RESULT_TOLERR   = 'testTolError'
    RESULT_INPUT    = 'testInput'
    RESULT_OUTPUT   = 'testOutput'
    RESULT_REF      = 'testReference'
    RESULT_IGNORED  = 'testResultIgnored'
    RESULT_INFO     = 'testInfo'
    RESULT_TYPE     = 'testResultType'
    RESULT_PROX     = 'testResultProximity'
    ALG             = 'algorithm'
    ALG_ARGS        = 'algorithmArgs'
    ALG_KWARGS      = 'algorithmKwargs'
    REFALG          = 'refAlgorithm'
    REFALG_ARGS     = 'refAlgorithmArgs'
    REFALG_KWARGS   = 'refAlgorithmKwargs'

    CHECK_PROXIMITY = 'checkProximity'
    CHECK_DATATYPE  = 'checkDataType'
    TYPE_EXPECTED   = 'typeExpected'
    TYPE_PROMOTION  = 'typePromotion'

################################################################################
##################################################  test implementations


################################################## _compareResults
def compareResults(test, query):

    def getOption(option, default):
        return query.get(option, test.get(option, default))

    arrOutput=query[TEST.RESULT_OUTPUT]
    arrReference=query[TEST.RESULT_REF]
    arrMatrix=test[TEST.REFERENCE]
    instance=query[TEST.INSTANCE]=test[TEST.INSTANCE]

    # get comparision options from test and query dicts. Priority has query
    minimalType=np.dtype(getOption(TEST.TYPE_PROMOTION, np.int8))
    expectedType=np.dtype(getOption(TEST.TYPE_EXPECTED,
                                    np.promote_types(instance.dtype,
                                                     minimalType)))
    maxEps=getOption(TEST.TOL_MINEPS, 0.)
    tolPower=getOption(TEST.TOL_POWER, 1.)
    ignoreType=not getOption(TEST.CHECK_DATATYPE, True)
    ignoreProximity=not getOption(TEST.CHECK_PROXIMITY, True)

    # check if shapes match. If they do, check all elements
    if arrOutput.shape != arrReference.shape:
        query[TEST.RESULT]=False
        query[TEST.RESULT_IGNORED]=False
        query[TEST.RESULT_INFO]=fmtRed(
            "%s!=%s" %(str(arrOutput.shape), str(arrReference.shape)))
        return query

    # if the query was not generated from input data, assume an int8-Eye-Matrix
    if TEST.RESULT_INPUT in query:
        arrInput=query[TEST.RESULT_INPUT]
        # check that input vector actually contains energy
        if np.linalg.norm(arrInput) == 0:
            print("Test query data causing exception:")
            pprint(query)
            raise ValueError("All-zero input vector detected in test")

        expectedType=np.promote_types(expectedType, arrInput.dtype)
        maxEps=max(maxEps, _getTypeEps(arrInput.dtype))

    # compare returned output data type to expected (qType) to verify
    # functionality of fastmat built-in type promotion mechanism
    resultType=np.can_cast(arrOutput.dtype, expectedType, casting='no')

    maxEps=max(maxEps,
               _getTypeEps(np.promote_types(minimalType, instance.dtype)),
               _getTypeEps(np.promote_types(minimalType, arrReference.dtype)))

    # determine allowed tolerance maxima (allow accuracy degradation of chained
    # operations by representing multiple stages by a power on operation count
    # the test distribution function generates random arrays with their absolute
    # element values in the range [0.4 .. 0.8]. This can be described by a
    # `dynamics`-factor of 2 per computation stage (parameter TEST.TOL_POWER)
    dynamics=2.
    maxDim=max(arrMatrix.shape + instance.shape)
    tolError=5 * dynamics * maxEps * (dynamics * np.sqrt(maxDim)) ** tolPower
    query[TEST.RESULT_TOLERR]=tolError

    maxRef = float(np.amax(np.abs(arrReference)))
    maxDiff = float(np.amax(np.abs(arrOutput - arrReference)))
    error = (maxDiff / maxRef if maxRef != 0 else maxDiff)
    resultProximity=(error <= tolError) or (maxRef <= tolError)

    # determine final result
    query[TEST.RESULT]=(resultType and resultProximity)
    # result ignored: whenever the main result is not true but an ignore in one
    # of the tests would cause it to become true
    query[TEST.RESULT_IGNORED]=((resultType or ignoreType) and
                                (resultProximity or ignoreProximity) and
                                not query[TEST.RESULT])
    query[TEST.RESULT_TYPE]=(resultType, ignoreType,
                             arrOutput.dtype, expectedType)
    query[TEST.RESULT_PROX]=(resultProximity, ignoreProximity, error, maxRef)

    return query


################################################## initTest()
def formatResult(result):
    # if a result info was already generated, return this. Otherwise generate it
    if TEST.RESULT_INFO in result:
        return result[TEST.RESULT_INFO]

    # fetch detailed results from result
    resultType, ignoreType, outputType, expectedType=result[TEST.RESULT_TYPE]
    resultProx, ignoreProx, error, maxRef=result[TEST.RESULT_PROX]

    # format string output
    if resultType:
        strInfo=fmtGreen(NAME.TYPENAME[outputType])
    elif ignoreType:
        strInfo=fmtYellow(NAME.TYPENAME[outputType])
    else:
        strInfo=fmtRed("%s!=%s"% (NAME.TYPENAME[outputType],
                                  NAME.TYPENAME[expectedType]))

    if not resultProx:
        if error > 0 and maxRef > 0:
            try:
                strDiff = "%+2.2d" %(
                    int(min(99, np.round(np.log10(1. * error / maxRef)))))
            except OverflowError:
                strDiff = '---'
        else:
            strDiff = '000'

        strInfo=strInfo + (fmtYellow if ignoreProx else fmtRed)(strDiff)

    return strInfo


################################################## initTest()
def initTest(test):
    # generate test object instance for given parameter set
    test[TEST.INSTANCE]=test[TEST.OBJECT](
        *test.get(TEST.INITARGS, ()),
        **test.get(TEST.INITKWARGS, {}))

    # generate plain reference array
    test[TEST.REFERENCE]=test[TEST.INSTANCE].reference()


################################################## testFailDump()
def testFailDump(test):
    print("Test query causing the exception:")
    pprint(test)


################################################## testArrays()
def testArrays(test):
    query={TEST.RESULT_INPUT      : test[TEST.RESULT_INPUT],
           TEST.RESULT_OUTPUT     : test[TEST.RESULT_OUTPUT],
           TEST.RESULT_REF        : test[TEST.RESULT_REF]}
    return compareResults(test, query)


################################################## testForward()
def testForward(test):
    query={}
    arrInput=query[TEST.RESULT_INPUT]=test[TEST.DATAARRAY].forwardData
    arrInputCheck = arrInput.copy()
    query[TEST.RESULT_OUTPUT]=test[TEST.INSTANCE].forward(arrInput)
    if not np.array_equal(arrInput, arrInputCheck):
        testFailDump(test)
        raise ValueError(".forward() modified input array.")

    query[TEST.RESULT_REF]=test[TEST.REFERENCE].dot(arrInput)
    return compareResults(test, query)


################################################## testBackward()
def testBackward(test):
    query={}
    arrInput=query[TEST.RESULT_INPUT]=test[TEST.DATAARRAY].backwardData
    arrInputCheck = arrInput.copy()
    query[TEST.RESULT_OUTPUT]=test[TEST.INSTANCE].backward(arrInput)
    if not np.array_equal(arrInput, arrInputCheck):
        testFailDump(test)
        raise ValueError(".backward() modified input array.")

    query[TEST.RESULT_REF]=test[TEST.REFERENCE].T.conj().dot(arrInput)
    return compareResults(test, query)


################################################## testArray()
def testArray(test):
    query={}
    instance=test[TEST.INSTANCE]
    query[TEST.RESULT_OUTPUT]=instance.array
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].astype(instance.dtype)
    return compareResults(test, query)


################################################## testGetItem()
def testGetItem(test):
    query={}
    instance=test[TEST.INSTANCE]
    arrOutput=np.zeros(instance.shape, instance.dtype)
    for nn, mm in itertools.product(range(instance.numN), range(instance.numM)):
        arrOutput[nn, mm]=instance[nn, mm]
    query[TEST.RESULT_OUTPUT]=arrOutput
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].astype(instance.dtype)
    return compareResults(test, query)


################################################## testGetColsSingle()
def testGetColsSingle(test):
    query={}
    instance=test[TEST.INSTANCE]
    arrOutput=np.zeros(instance.shape, instance.dtype)
    for mm in range(instance.numM):
        arrOutput[:, mm]=instance.getCols(mm)
    query[TEST.RESULT_OUTPUT]=arrOutput
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].astype(instance.dtype)
    return compareResults(test, query)


################################################## testGetColsMultiple()
def testGetColsMultiple(test):
    query={}
    instance=test[TEST.INSTANCE]
    arrOutput=instance.getCols([c for c in range(instance.numM)])
    query[TEST.RESULT_OUTPUT]=arrOutput
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].astype(instance.dtype)
    return compareResults(test, query)


################################################## test: getRowsSingle
def testGetRowsSingle(test):
    query={}
    instance=test[TEST.INSTANCE]
    arrOutput=np.zeros((instance.numN, instance.numM), instance.dtype)
    for nn in range(instance.numN):
        arrOutput[nn, :]=instance.getRows(nn)
    query[TEST.RESULT_OUTPUT]=arrOutput
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].astype(instance.dtype)
    return compareResults(test, query)


################################################## test: getRowsMultiple
def testGetRowsMultiple(test):
    query={}
    instance=test[TEST.INSTANCE]
    arrOutput=instance.getRows([r for r in range(instance.numN)])
    query[TEST.RESULT_OUTPUT]=arrOutput
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].astype(instance.dtype)
    return compareResults(test, query)


################################################## test: normalized (property)
def testNormalized(test):
    instance, reference=test[TEST.INSTANCE], test[TEST.REFERENCE]

    # usually expect the normalized matrix to be promoted in type complexity
    # due to division by column-norm during the process. However there exist
    # matrices that treat the problem differently. Exclude the expected pro-
    # motion for them.
    query=({} if isinstance(instance, (Diag, Eye, Zero))
           else {TEST.TYPE_PROMOTION: np.float32})

    # ignore actual type of generated gram:
    query[TEST.CHECK_DATATYPE]=False
    query[TEST.TOL_MINEPS]=_getTypeEps(safeTypeExpansion(instance.dtype))

    try:
        query[TEST.RESULT_OUTPUT]=instance.normalized.array
        query[TEST.RESULT_REF]=np.einsum(
            'ij,j->ij', reference,
            1. / np.apply_along_axis(np.linalg.norm, 0, reference))
        return compareResults(test, query)
    except ValueError:
        if isinstance(instance, Zero):
            # failing normalization is expected for Zero matrix.
            result, ignored=True, False
        elif isinstance(instance, Parametric):
            # Parametric normalization is excused for now (int8 trouble)
            result, ignored=False, True
        else:
            result, ignored=False, False

        query[TEST.RESULT], query[TEST.RESULT_IGNORED]=result, ignored
        query[TEST.RESULT_INFO]='!RNK'
        return query


################################################## test: largestSV (property)
def testLargestSV(test):
    query={TEST.TYPE_EXPECTED: np.float64}
    instance=test[TEST.INSTANCE]

    # account for "extra computation stage" (gram) in largestSV
    query[TEST.TOL_POWER]=test.get(TEST.TOL_POWER, 1.) * 2
    query[TEST.TOL_MINEPS]=_getTypeEps(safeTypeExpansion(instance.dtype))

    # determine reference result
    largestSV=np.linalg.svd(test[TEST.REFERENCE], compute_uv=False)[0]
    query[TEST.RESULT_REF]=np.array(
        largestSV, dtype=np.promote_types(largestSV.dtype, np.float64))

    # largestSV may not converge fast enough for a bad random starting point
    # so retry some times before throwing up
    for tries in range(9):
        maxSteps=100. * 10. ** (tries / 2.)
        query[TEST.RESULT_OUTPUT]=np.array(
            instance.getLargestSV(maxSteps=maxSteps, alwaysReturn=True))
        result=compareResults(test, query)
        if result[TEST.RESULT]:
            break
    return result


################################################## test: gram (property)
def testGram(test):
    instance, reference=test[TEST.INSTANCE], test[TEST.REFERENCE]

    # usually expect the normalized matrix to be promoted in type complexity
    # due to division by column-norm during the process. However there exist
    # matrices that treat the problem differently. Exclude the expected pro-
    # motion for them.
    query=({} if isinstance(instance, (Diag, Eye, Zero))
           else {TEST.TYPE_PROMOTION: np.float32})

    # account for "extra computation stage" in gram
    query[TEST.TOL_POWER]=test.get(TEST.TOL_POWER, 1.) * 2

    query[TEST.RESULT_OUTPUT]=instance.gram.array
    query[TEST.RESULT_REF]=reference.astype(
        np.promote_types(np.float32, reference.dtype)).T.conj().dot(reference)

    # ignore actual type of generated gram:
    query[TEST.CHECK_DATATYPE]=False

    return compareResults(test, query)


################################################## test: T (property)
def testTranspose(test):
    query={}
    instance=test[TEST.INSTANCE]
    query[TEST.RESULT_OUTPUT]=instance.T.array
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].T
    return compareResults(test, query)


################################################## test: H (property)
def testHermitian(test):
    query={}
    instance=test[TEST.INSTANCE]
    query[TEST.RESULT_OUTPUT]=instance.H.array
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].T.conj()
    return compareResults(test, query)


################################################## test: conj (property)
def testConjugate(test):
    query={}
    instance=test[TEST.INSTANCE]
    query[TEST.RESULT_OUTPUT]=instance.conj.array
    query[TEST.RESULT_REF]=test[TEST.REFERENCE].conj()
    return compareResults(test, query)


################################################## test: Algorithm
def testAlgorithm(test):
    # generate input data vector. As this one is needed for some algs, do one
    # dereferentiation step on a complete dictionary, including test and result
    query=test.copy()
    arrInput, query[TEST.RESULT_INPUT]=query[TEST.DATAARRAY].forwardData
    paramDereferentiate(query)

    # now call the algorithm executor functions for FUT and reference
    arrInputCheck = arrInput.copy()
    query[TEST.RESULT_OUTPUT]=query[TEST.ALG](
        *query.get(TEST.ALG_ARGS, ()),
        **query.get(TEST.ALG_KWARGS, {}))
    query[TEST.RESULT_REF]=query[TEST.REFALG](
        *query.get(TEST.REFALG_ARGS,      query.get(TEST.ALG_ARGS, [])),
        **query.get(TEST.REFALG_KWARGS,   query.get(TEST.ALG_KWARGS, {})))

    if not np.array_equal(arrInput, arrInputCheck):
        testFailDump(test)
        raise ValueError("Algorithm modified input array.")

    return compareResults(test, query)


################################################################################
################################################## class TestRunner
class Test(Worker):

    _verboseFull=False

    def __init__(self, targetClass, **options):

        # extract options
        extraOptions = options.get('extraOptions',  {})

        # by default, enable verbosity for issues and isolate problems
        self.cbStatus=self.printStatus
        self.cbResult=self.findProblems

        # initialize output collectors
        self.problems = AccessDict({})
        self.irregularities = AccessDict({})

        # define test-specific defaults
        defaults={
            TEST.COMMON: {
                NAME.NAME       : dynFormat("%s.%s", NAME.CLASS, NAME.TARGET),
                TEST.NAMINGARGS : dynFormat("[%dx%d]", TEST.NUM_N, TEST.NUM_M),
                TEST.NAMING     : dynFormat(
                    "%s(%s)", NAME.CLASS, TEST.NAMINGARGS)

            }, TEST.CLASS: {
                # define tests
                TEST.QUERY: IgnoreDict({
                    'arr'   : testArray,
                    'item'  : testGetItem,
                    'gCs'   : testGetColsSingle,
                    'gCm'   : testGetColsMultiple,
                    'gRs'   : testGetRowsSingle,
                    'gRm'   : testGetRowsMultiple,
                    'nor'   : testNormalized,
                    'lSV'   : testLargestSV,
                    'gram'  : testGram,
                    'T'     : testTranspose,
                    'H'     : testHermitian,
                    'conj'  : testConjugate
                })

            }, TEST.TRANSFORMS: {
                # define default naming and captions
                TEST.NAMING     : dynFormat(
                    "%s(%s)*%s", NAME.CLASS, TEST.NAMINGARGS, TEST.DATAARRAY),

                # define default arrB ArrayGenerator
                TEST.DATACOLS   : Permutation([1, 5]),
                TEST.DATATYPE   : VariantPermutation(TEST.ALLTYPES),
                TEST.DATASHAPE  : (TEST.NUM_M, TEST.DATACOLS),
                TEST.DATASHAPE_T: (TEST.NUM_N, TEST.DATACOLS),
                TEST.DATAALIGN  : VariantPermutation(TEST.ALLALIGNMENTS),
                TEST.DATACENTER : 0,
                TEST.DATAARRAY  : ArrayGenerator({
                    NAME.DTYPE      : TEST.DATATYPE,
                    NAME.SHAPE      : TEST.DATASHAPE,
                    NAME.SHAPE_T    : TEST.DATASHAPE_T,
                    NAME.ALIGN      : TEST.DATAALIGN,
                    NAME.CENTER     : TEST.DATACENTER
                }),

                # define tests
                TEST.QUERY      : IgnoreDict({
                    'F'     : testForward,
                    'B'     : testBackward,
                })

            }, TEST.ALGORITHM: {
                # define default naming and captions
                TEST.TOL_POWER  : 1.,
                TEST.NAMING     : dynFormat(
                    "%s(%s)*%s", NAME.CLASS, TEST.NAMINGARGS, TEST.DATAARRAY),

                # define default arrB ArrayGenerator
                TEST.DATACOLS   : Permutation([1, 5]),
                TEST.DATATYPE   : VariantPermutation(TEST.LARGETYPES),
                TEST.DATASHAPE  : (TEST.NUM_M, TEST.DATACOLS),
                TEST.DATASHAPE_T: (TEST.NUM_N, TEST.DATACOLS),
                TEST.DATAALIGN  : VariantPermutation(TEST.ALLALIGNMENTS),
                TEST.DATACENTER : 0,
                TEST.DATAARRAY  : ArrayGenerator({
                    NAME.DTYPE      : TEST.DATATYPE,
                    NAME.SHAPE      : TEST.DATASHAPE,
                    NAME.SHAPE_T    : TEST.DATASHAPE_T,
                    NAME.ALIGN      : TEST.DATAALIGN,
                    NAME.CENTER     : TEST.DATACENTER
                }),

                # define tests
                TEST.QUERY: IgnoreDict({
                    'arr'     : testArrays
                })

            }
        }

        # call parent initialization with Test-specific options
        super(Test, self).__init__(
            targetClass, targetOptionMethod='_getTest',
            runnerDefaults=defaults, extraOptions=extraOptions)

    def _runTest(self, name, options):

        # build list of tests as complete permutation of parameter variations
        tests=uniqueNameDict({})
        lstTests=paramPermute(options)
        for test in lstTests:
            # prepare test dictionary (dereferentiate links, evaluate functions)
            paramDereferentiate(test)
            paramEvaluate(test)

            # name each test instance from the key naming rule option ('naming')
            # 'naming' is a format string, which selects its fields from the
            # option dictionary made up by the target dictionary itself.
            tests[getattr(test, TEST.NAMING, test[NAME.TARGET])]=test

            # sanity-check proper definition of test target
            obj=test.get(TEST.OBJECT, None)
            if (obj is None or
                    not (isinstance(obj, IgnoreFunc) or
                         issubclass(obj, Matrix))):
                raise ValueError("%s['%s'] not a fastmat class or test case." %(
                    test.name, TEST.OBJECT))

        # When done, convert tests back to regular dictionary, enabling editing
        tests=dict(tests)

        # allow a second level of Permutation for VariantPermutation instances
        # determine variants of target and generate variant description tag name
        lstVariantPermutations=[name for name, value in options.items()
                                if isinstance(value, VariantPermutation)]
        descrVariants=','.join(lstVariantPermutations)

        # determine field lengths for nice printing of columns
        lenName=max(map(len, tests.keys()))

        # self.results is constructed as multi-level dictionary:
        #     ~[target-name][test-name][variant-name][query-name]: results dict
        # initialize result structure for [test-name] level
        # create the [variant-level] dictionary as uniqueNameDict to prevent
        # overwriting existing variant results with the same name
        resultTarget={name: uniqueNameDict({}) for name in tests.keys()}

        # iterate through tests
        for nameTest, test in sorted(tests.items()):
            # initialize instances and required stuff for each test
            initTest(test)

            # get a pointer to the test result dictionary
            resultTest=resultTarget[nameTest]

            lstVariants=paramPermute(
                test, PermutationClass=VariantPermutation)
            variants=[paramDereferentiate(variant) for variant in lstVariants]
            for variant in variants:
                variant[NAME.VARIANT]=''.join(
                    [NAME.TYPENAME[variant[key]]
                     if isinstance(variant[key], type)
                     else str(variant[key])
                     for key in lstVariantPermutations])

                # call variant initialization routine, if defined
                if TEST.INIT_VARIANT in variant:
                    variant[TEST.INIT_VARIANT](variant)

                # execute test queries, collect results as [query-name] level
                # and store in [variant-name] level into test result structure.
                resultTest[variant[NAME.VARIANT]]={
                    name: query(variant)
                    for name, query in test[TEST.QUERY].items()
                }

            self.emitStatus(nameTest, resultTest, lenName, descrVariants)

        return resultTarget

    def _run(self, name, options):

        maxTries = 3
        for numTry in range(maxTries):
            resultTarget = self._runTest(name, options.copy())

            result = all(all(all(resultQuery[TEST.RESULT]
                                 for resultQuery in resultVariant.values())
                             for resultVariant in resultTest.values())
                         for resultTest in resultTarget.values())

            if result:
                break
            else:
                print("Test %s.%s failed in during try #%d/%d.%s" %(
                    options[NAME.CLASS], name, numTry + 1, maxTries,
                    " Retrying ..." if numTry < maxTries - 1 else ""))

        return resultTarget

    def printStatus(self, nameTest, resultTest, lenName=-1, descrVariants=""):
        # if full output isn't actually requested, crawl all query results in
        # the hierarchy of this test for negative results. If none are found,
        # skip printing.
        if not self._verboseFull:
            if all(all(query.get(TEST.RESULT, True)
                       for query in variant.values())
                   for variant in resultTest.values()):
                return

        lenNameV=max(len(nameV) for nameV in resultTest.keys())
        if lenName < 1:
            lenName=len(nameTest)

        # construct named prefix describing the test and its following variants
        strPrefix="%-*s " %(lenName + len(descrVariants),
                            nameTest + ":" + descrVariants)

        # convert results to info strings
        strQueries={nameV: ["%s:%s" %(nameQ, formatResult(query))
                            for nameQ, query in sorted(variant.items())]
                    for nameV, variant in resultTest.items()}

        # determine print-lengths of query strings
        lenQueries={nameV: [len(fmtEscape(strQuery)) for strQuery in variant]
                    for nameV, variant in strQueries.items()}

        # determine highest print-length for any query
        maxLenQuery=max(max(variant) for variant in lenQueries.values())

        # normalize string lengths in strQueries
        for nameV, lengths in lenQueries.items():
            strQueries[nameV]=["%s%s" %(strQueries[nameV][qq],
                                        " " * (maxLenQuery - lenQuery))
                               for qq, lenQuery in enumerate(lengths)]

        # join query strings to form variant strings
        strVariants=["%*s %s" %(lenNameV, nameV, " ".join(lstStr))
                     for nameV, lstStr in sorted(strQueries.items())]

        lenVariants=[len(fmtEscape(strV)) for strV in strVariants]
        maxLenVariants=max(lenVariants)

        # normalize string lengths in strVariants
        strVariants=["%s%s" %(strV, " " * (maxLenVariants - lenVariants[vv]))
                     for vv, strV in enumerate(strVariants)]
        lenVariants=(0 if len(strVariants) == 0
                     else len(fmtEscape(strVariants[0])))

        # calculate amount of entries printable per line
        width=self.consoleWidth
        itemsPerLine=(
            1 if width < 0
            else max(1, ((width - len(strPrefix)) // (1 + lenVariants))))

        for ii in range(0, len(strVariants), itemsPerLine):
            print("%s|%s" %(strPrefix if ii == 0 else " " * len(strPrefix),
                            " |".join(strVariants[ii: ii + itemsPerLine])))

    # control verbosity of _run() and printStatus()
    @property
    def verbosity(self):
        return self.cbStatus == self.printStatus, self._verboseFull

    @verbosity.setter
    def verbosity(self, value):
        if isinstance(value, bool):
            value=(value, )

        if isinstance(value, tuple):
            if len(value) >= 1:
                self.cbStatus=(self.printStatus if value[0] is True
                               else None)
            if len(value) >= 2:
                self._verboseFull=value[1]

    def findProblems(self, nameTarget, targetResult):
        # crawl all query results for problems and irregularities. Collect them
        # if found. Also, construct a considerable name.
        problems={}
        irregularities={}
        for nameT, test in sorted(targetResult.items()):
            for nameV, variant in sorted(test.items()):
                for nameQ, query in sorted(variant.items()):
                    if not query.get(TEST.RESULT, False):
                        nameTestQuery="%s%s.%s" %(nameT, nameV, nameQ)
                        if query.get(TEST.RESULT_IGNORED, False):
                            irregularities[nameTestQuery]=query
                        else:
                            problems[nameTestQuery]=query

        self.problems[nameTarget] = convertToAccessDicts(problems)
        self.irregularities[nameTarget] = convertToAccessDicts(irregularities)

        return self.problems[nameTarget]



# -*- coding: utf-8 -*-
'''
  fastmat/LowRank.pxd
 -------------------------------------------------- part of the fastmat package

  Header file for LowRank base class (structural description).


  Author      : sempersn
  Introduced  : 2017-02-02
 ------------------------------------------------------------------------------

   Copyright 2017 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import cython
cimport numpy as np

from .Product cimport Product

################################################################################
################################################## class LowRank
cdef class LowRank(Product):

    ############################################## class variables
    cdef public np.ndarray _vecS                 # matrix singular values
    cdef public np.ndarray _arrU                 # left orthogonal matrix
    cdef public np.ndarray _arrV                 # right orthogonal matrix
    cdef public list _summands                   # list of generated rank-1
    #                                            # summands

    ############################################## class methods
    cpdef np.ndarray _reference(self)



# -*- coding: utf-8 -*-
'''
  fastmat/algs/FISTA.py
 -------------------------------------------------- part of the fastmat package

  Implementation of Fast Iterative Shrinking-Thresholding Algorithm (FISTA) in
  fastmat.


  Author      : kirchhjn
  Introduced  : 2017-11-06
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------

  TODO:
    - Todos for ISTA
    - Check if its working
'''
import numpy as np

from ..base import Algorithm


################################################################################
###  FISTA: Sparse recovery algorithm
################################################################################


def _softThreshold(arrX, numAlpha):
    '''
    Do a soft Thresholding step.
      arrM         - positive part of arrX - numAlpha
      arrX         - vector to be thresholded
      numAlpha     - thresholding threshold
    '''
    arrM = np.maximum(np.abs(arrX) - numAlpha, 0)
    return np.multiply((arrM / (arrM + numAlpha)), arrX)


##################################################

def FISTA(
        fmatA,
        arrB,
        numLambda=0.1,
        numMaxSteps=100
):
    '''
    Wrapper around the FISTA algrithm to allow processing of arrays of signals
        fmatA         - input system matrix
        arrB          - input data vector (measurements)
        numLambda     - balancing parameter in optimization problem
                        between data fidelity and sparsity
        numMaxSteps   - maximum number of steps to run
        numL          - step size during the conjugate gradient step
    '''

    if len(arrB.shape) > 2:
        raise ValueError("Only n x m arrays are supported for FISTA")

    # calculate the largest singular value to get the right step size
    numL = 1.0 / (fmatA.largestSV ** 2)
    t = 1
    arrX = np.zeros(
        (fmatA.numM, arrB.shape[1]),
        dtype=np.promote_types(np.float32, arrB.dtype)
    )
    # initial arrY
    arrY = np.copy(arrX)
    # start iterating
    for numStep in range(numMaxSteps):
        arrXold = np.copy(arrX)
        # do the gradient step and threshold
        arrStep = arrY - numL * fmatA.backward(fmatA.forward(arrY) - arrB)

        arrX = _softThreshold(arrStep, numL * numLambda * 0.5)

        # update t
        tOld =t
        t = (1 + np.sqrt(1 + 4 * t ** 2)) / 2
        # update arrY
        arrY = arrX + ((tOld - 1) / t) * (arrX - arrXold)
    # return the unthresholded values for all non-zero support elements
    return np.where(arrX != 0, arrStep, arrX)


################################################################################
###  Maintenance and Documentation
################################################################################

################################################## inspection interface
class FISTAinspect(Algorithm):
    @staticmethod
    def _getTest():
        from ..inspect import TEST, dynFormat, arrSparseTestDist
        from ..core.types import _getTypeEps
        from ..Product import Product
        from ..Hadamard import Hadamard
        from ..Matrix import Matrix

        def testFISTA(test):
            # prepare vectors
            numM = test[TEST.NUM_M]
            test[TEST.RESULT_REF] = np.hstack(
                [arrSparseTestDist((numM, 1), dtype=test[TEST.DATATYPE],
                                   density=1. * test['numK'] / numM).todense()
                 for nn in range(test[TEST.DATACOLS])])
            test[TEST.RESULT_INPUT] = (test[TEST.INSTANCE] *
                                       test[TEST.RESULT_REF])
            test[TEST.RESULT_OUTPUT] = FISTA(
                test[TEST.INSTANCE], test[TEST.RESULT_INPUT],
                numLambda=test['lambda'], numMaxSteps=test['maxSteps'])

        return {
            TEST.ALGORITHM: {
                'order': 6,
                TEST.NUM_N: (lambda param: 3 * param['order']),
                TEST.NUM_M: (lambda param: 2 ** param['order']),
                'numK': 'order',
                'lambda': 10.,
                'maxSteps': 1000,
                'typeA': TEST.Permutation(TEST.ALLTYPES),

                TEST.OBJECT: Matrix,
                TEST.INITARGS: (lambda param: [
                    Product(Matrix(np.random.uniform(
                        -100, 100, (getattr(param, TEST.NUM_M),
                                    getattr(param, TEST.NUM_M))).astype(
                        param['typeA'])),
                        Hadamard(param.order),
                        typeExpansion=param['typeA']).array]),

                TEST.DATAALIGN: TEST.ALIGNMENT.DONTCARE,
                TEST.INIT_VARIANT: TEST.IgnoreFunc(testFISTA),

                'strTypeA': (lambda param: TEST.TYPENAME[param['typeA']]),
                TEST.NAMINGARGS: dynFormat("(%dx%d)*Hadamard(%s)[%s]",
                                           TEST.NUM_N, TEST.NUM_M,
                                           'order', 'strTypeA'),

                # matrix inversion always expands data type to floating-point
                TEST.TYPE_PROMOTION: np.float32,
                TEST.TOL_MINEPS: _getTypeEps(np.float32),
                TEST.TOL_POWER: 5.
                # TEST.CHECK_PROXIMITY    : False
            },
        }

    @staticmethod
    def _getBenchmark():
        from ..inspect import BENCH, arrTestDist
        from ..Matrix import Matrix
        from ..Product import Product
        from ..Fourier import Fourier
        from scipy import sparse as sps

        def createTarget(M, datatype):
            '''Create test target for algorithm performance evaluation.'''

            if M < 10:
                raise ValueError("Problem size too small for FISTA benchmark")

            # assume a 1:5 ratio of measurements and problem size
            # assume a sparsity of half the number of measurements
            N = int(np.round(M / 5.0))
            K = int(N / 2)

            # generate matA (random measurement matrix, Fourier dictionary)
            matA = Product(Matrix(arrTestDist((N, M), datatype)), Fourier(M))

            # generate arrB from random baseline support (RHS)
            arrB = matA * sps.rand(M, 1, 1.0 * K / M).todense().astype(datatype)

            return (FISTA, [matA, arrB])

        return {
            BENCH.COMMON: {
                BENCH.NAME: 'FISTA Algorithm',
                BENCH.FUNC_GEN: (lambda c: createTarget(10 * c, np.float64)),
                BENCH.FUNC_SIZE: (lambda c: 10 * c)
            },
            BENCH.PERFORMANCE: {
                BENCH.CAPTION: 'FISTA performance'
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN: (lambda c, dt: createTarget(10 * c, dt)),
                BENCH.FUNC_SIZE: (lambda c: 10 * c),
                BENCH.FUNC_STEP: (lambda c: c * 10 ** (1. / 12)),
            }
        }

    @staticmethod
    def _getDocumentation():
        from ..inspect import DOC
        return DOC.SUBSECTION(
            r"""
Fast Iterative Soft Thresholding Algorithm (FISTA)
(\texttt{fastmat.algs.FISTA})""",
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
For a given matrix $\bm A \in \C^{m \times N}$ with $m \ll N$ and a
vector $\bm b \in \C^m$ we approximately solve

\begin{align}\label{fista_lasso}
    \Min\limits_{\bm x \in \C^N}\Norm{\bm A \cdot \bm x - \bm b}^2_2 +
    \lambda\cdot\Norm{\bm x}_1,
\end{align}

where $\lambda > 0$ is a regularization parameter to steer the trade-off between
data fidelity and sparsity of the solution.

\begin{itemize}
\item \textbf{Input:} Full rank matrix $\bm A$, measurements $\bm b$, trade-off
    parameter $\lambda$
\item \textbf{Output:} Reconstruction vector $\bm x$.
\end{itemize}

FISTA is a performance upgrade for the standard ISTA algorithm as described
 in\cite{fista_beck1952fista}.""",
                DOC.SNIPPET('# import the packages',
                            'import numpy.linalg as npl',
                            'import numpy as np',
                            'import fastmat as fm',
                            'import fastmat.algs as fma',
                            '',
                            '# define the dimensions',
                            '# and the sparsity',
                            'n, k = 512, 3',
                            '',
                            '# define the sampling positions',
                            't = np.linspace(0, 20 * np.pi, n)',
                            '',
                            '# construct the convolution matrix',
                            'c = np.cos(2 * t),',
                            'C = fm.Circulant(c)',
                            '',
                            '# create the ground truth',
                            'x = np.zeros(n)',
                            'x[npr.choice(range(n),',
                            '             k, replace=0)] = 1',
                            'b = C * x',
                            '',
                            '# reconstruct it',
                            'y = fma.FISTA(C, b, 0.005, 1000)',
                            '',
                            '# test if they are close in the',
                            '# domain of C',
                            'print(npl.norm(C * y - b))',
                            caption=r"""
We solve a sparse deconvolution problem, where the atoms are harmonics windowed
by a gaussian envelope. The ground truth $\bm x$ is build out of three pulses at
arbitrary locations."""),
                r"""
\textit{Hint:} The proper choice of $\lambda$ is crucial for good perfomance of
    this algorithm, but this is not an easy task. Unfortunately we are not
    in the place here to give you a rule of thumb what to do, since it
    highly depends on the application at hand. Again, consult
    \cite{fista_beck1952fista} for any further considerations of this matter."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
We use $\bm A = \bm M \cdot \bm \Fs$, where $\bm M$ was drawn from a standard
Gaussian distribution and $\bm \Fs$ is a Fourier matrix. The vector $\bm b \in
\C^m$ of equation \eqref{fista_lasso} is generated from multiplying $\bm A$ with
a sparse vector $\bm x$.""",
                DOC.PLOTPERFORMANCE(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            ),
            DOC.BIBLIO(
                fista_beck1952fista=DOC.BIBITEM(
                    r'Amir Beck and Marc Teboulle',
                    r"""
A Fast Iterative Shrinkage-Thresholding Algorithm for Linear Inverse Problems
""",
                    r'SIAM Journal on Imaging Sciences 2009 2:1, 183-202')
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/Polynomial.py
 -------------------------------------------------- part of the fastmat package

  Determine the Polynomial of a fastmat matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .core.types cimport *
from .Matrix cimport Matrix


################################################################################
################################################## class Polynomial
cdef class Polynomial(Matrix):

    ############################################## class properties
    # coeff - Property (read-only)
    # Return the polynomial coefficient vector.
    property coeff:
        def __get__(self):
            return self._coeff

    ############################################## class methods
    def __init__(self, mat, coeff, **options):
        '''Initialize Matrix instance'''

        if mat.numN != mat.numM:
            raise ValueError("Polynomial: Matrix must be square.")

        dtype = np.promote_types(mat.dtype, coeff.dtype)

        # handle type expansion with default depending on matrix type
        # default: expand small types due to accumulation during transforms
        # skip by specifying `typeExpansion=None` or override with `~=...`
        typeExpansion = options.get('typeExpansion', safeTypeExpansion(dtype))
        dtype = (dtype if typeExpansion is None
                 else np.promote_types(dtype, typeExpansion))

        self._content = (mat,)
        self._coeff = np.flipud(coeff.astype(dtype))
        self._coeffConj = self._coeff.conj()

        # set properties of matrix
        self._initProperties(self._content[0].numN, self._content[0].numM,
                             dtype)

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = len(self._coeff) * (self.numN + self.numM)
        return (complexity, complexity)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Apply the Horner scheme to a polynomial of matrices.'''
        cdef cc, cnt = self._coeff.shape[0]
        cdef np.ndarray arrRes, arrIn = arrX

        arrRes = np.inner(arrX, self._coeff[0])

        # use inner for element-wise scalar mul as inner does type promotion
        for cc in range(1, cnt):
            arrRes  = self._content[0].forward(arrRes) + np.inner(
                arrX, self._coeff[cc])

        return arrRes

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Apply the Horner scheme to a polynomial of matrices.'''
        cdef cc, cnt = self._coeffConj.shape[0]
        cdef np.ndarray arrRes

        arrRes = np.inner(arrX, self._coeffConj[0])

        # use inner for element-wise scalar mul as inner does type promotion
        for cc in range(1, cnt):
            arrRes  = self._content[0].backward(arrRes) + np.inner(
                arrX, self._coeffConj[cc])

        return arrRes

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        cdef intsize ii, ind = 0
        cdef np.ndarray arrRes, tmp

        dtype = np.promote_types(self.dtype, np.float64)
        arrRes = np.zeros((self.numN, self.numN), dtype=dtype)

        for cc in np.flipud(self._coeff):
            arrTrafo = self._content[0].reference()
            tmp = np.eye(self._content[0].numN, dtype=dtype)
            for ii in range(ind):
                tmp = arrTrafo.dot(tmp)

            arrRes = arrRes + np.inner(tmp, cc)
            ind += 1

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                'order'         : 5,
                TEST.TOL_POWER  : 'order',
                TEST.NUM_N      : 7,
                TEST.NUM_M      : 7,

                'mTypeC'        : TEST.Permutation(TEST.ALLTYPES),
                'mTypeM'        : TEST.Permutation(TEST.ALLTYPES),
                TEST.PARAMALIGN : TEST.Permutation(TEST.ALLALIGNMENTS),
                'vecC'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeC',
                    TEST.SHAPE  : ('order', ),
                    TEST.ALIGN  : TEST.PARAMALIGN
                }),
                'arrM'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'mTypeM',
                    TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_M)
                }),

                TEST.OBJECT     : Polynomial,
                TEST.INITARGS   : (lambda param: [Matrix(param['arrM']()),
                                                  param['vecC']()]),
                TEST.NAMINGARGS : dynFormat("%s,%s", 'vecC', 'arrM'),
                TEST.TOL_POWER  : 'order'
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        from .Eye import Eye
        from .Hadamard import Hadamard
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: Polynomial(
                    Hadamard(c), np.random.uniform(1, 2, 6))),
                BENCH.FUNC_SIZE : (lambda c: 2 ** c),
                BENCH.FUNC_STEP : (lambda c: c + 1)
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN  : (lambda c: Polynomial(
                    Eye(2 ** c), np.random.uniform(1, 2, 10)))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Polynomial (\texttt{fastmat.Polynomial})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
For given coefficients $a_k,\dots,a_0 \in \C$ and a linear mapping $\bm A \in
\C^{n \times n}$, we define
\[\bm M = a_n \bm A^n + a_{n-1} \bm A^{n-1} + a_1 \bm A + a_0 \bm I.\]
The transform $\bm M \cdot \bm x$ can be calculated efficiently with Horner's
method.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the transforms',
                            'H = fm.Hadamard(n)',
                            '',
                            '# define the coefficient array',
                            'arr_a = [1, 2 + 1j, -3.0, 0.0]',
                            '',
                            '# define the polynomial',
                            'M = fm.Polynomial(H, arr_a)',
                            caption=r"""
Let $\bm H_n$ be the Hadamard matrix of order $n$. And let
$\bm a = (1, 2 + i, -3, 0) \in \C^{4}$ be a coefficient vector,
then the polynomial is defined as
\[\bm M = \bm H_n^3 + (2+i) \bm H_n^2 - 3 \bm H_n.\]""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
The \emph{forward} and \emph{solve} benchmarks were performed on a matrix
$\bm P = a_2 \cdot \bm \Hs_k^2 + a_1 \cdot \bm \Hs_k^1 + a_0 \cdot \bm I_{2^k}$
""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(doc=r"""
Polynomial of Identity $\bm I_{2^k}$ matrices with degree $10$""")
            )
        )
